"""System configurator page."""

from typing import Optional

import pandas as pd
import streamlit as st

import adijif
from adijif.clocks import supported_parts as csp
from adijif.converters import supported_parts as xsp

# from adijif.utils import get_jesd_mode_from_params
from ..utils import Page
from .helpers.datapath import gen_datapath

# Clocks
options_to_skip = ["list_references_available", "d_syspulse"]
parts_to_ignore = ["ad9545", "ad9523_1"]
sp = [p for p in csp if p not in parts_to_ignore]
# Put HMC7044 at the front
sp = [p for p in sp if p != "hmc7044"]
sp.insert(0, "hmc7044")

# Converters


class SystemConfigurator(Page):
    """System configurator tool page."""

    def __init__(self, state: Optional[object]) -> None:
        """Initialize system configurator page.

        Args:
            state: Application state object
        """
        self.state = state

    def write(self) -> None:
        """Render the system configurator page."""
        st.title("System Configurator")

        with st.expander("System Settings", expanded=True):
            hsx = st.selectbox(
                label="Select a converter part",
                options=xsp,
                format_func=lambda x: x.upper(),
                key="converter_part_select",
            )

            clock = st.selectbox(
                label="Select a clock part",
                options=sp,
                format_func=lambda x: x.upper(),
                key="clock_part_select",
            )

            fpga_kit = st.selectbox(
                label="Select an FPGA development kit",
                options=adijif.xilinx._available_dev_kit_names,
                format_func=lambda x: x.upper(),
                key="fpga_dev_kit_select",
            )

            reference_rate = st.number_input(
                f"Reference Rate (VCXO) for {clock.upper()} (Hz)",
                value=125000000,
                min_value=int(100e6),
                max_value=int(400e6),
            )

        sys = adijif.system(hsx.lower(), clock.lower(), "xilinx", reference_rate)

        # Get Converter clocking requirements
        sys.Debug_Solver = False

        # Get FPGA clocking requirements
        sys.fpga.setup_by_dev_kit_name(fpga_kit.lower())

        converter_c, fpga_c = st.columns(2)

        with converter_c:
            # st.header("Converter Configuration")
            with st.expander("Converter Settings", expanded=True):

                # Units GHz, MHz, kHz
                units = st.selectbox(
                    label="Select units for Converter Clock",
                    options=["Hz", "kHz", "MHz", "GHz"],
                    index=2,
                )
                if units == "Hz":
                    multiplier = 1
                elif units == "kHz":
                    multiplier = 1e3
                elif units == "MHz":
                    multiplier = 1e6
                elif units == "GHz":
                    multiplier = 1e9

                # Converter Clock
                converter_clock = st.number_input(
                    f"Converter Clock ({units})",
                    value=1e9 / multiplier,
                    format="%f",
                    min_value=1e6 / multiplier,
                    max_value=20e9 / multiplier,
                )
                # sys.converter.decimation = 1

                decimation = gen_datapath(sys.converter)
                sys.converter.sample_clock = converter_clock * multiplier / decimation

                # JESD modes
                qsm = sys.converter.quick_configuration_modes

                # Flatten dict for display as a list
                qsm_flat = {}
                for jesdclasses in qsm:
                    for mode in qsm[jesdclasses]:
                        settings = qsm[jesdclasses][mode]
                        other_settings = (
                            f" (M={settings['M']}, "
                            + f"F={settings['F']},"
                            + f"Np={settings['Np']}, "
                            + f"L={settings['L']}, S={settings['S']})"
                        )
                        qsm_flat[
                            f"{jesdclasses.upper()} Mode: {mode} {other_settings}"
                        ] = {
                            "mode": mode,
                            "jesdclass": jesdclasses,
                        }

                # Sort by mode
                qsm_flat = dict(
                    sorted(qsm_flat.items(), key=lambda item: item[1]["mode"])
                )

                mode = st.selectbox(
                    label="Select JESD Configuration Mode",
                    options=list(qsm_flat.keys()),
                    # options=list(qsm_flat.keys()),
                    # format_func=lambda x: f"{x} : {qsm_flat[x]}",
                )
                sys.converter.set_quick_configuration_mode(
                    qsm_flat[mode]["mode"], qsm_flat[mode]["jesdclass"]
                )

        # with clock_c:
        #     st.header("Clock Configuration")

        #     with st.container(border=True):
        #         st.markdown(cfg["clock"])

        with fpga_c:
            with st.expander("FPGA Settings", expanded=True):

                # sys.fpga.ref_clock_constraint = "Unconstrained"
                ref_clock_constraint = st.selectbox(
                    options=adijif.xilinx._ref_clock_constraint_options,
                    label="FPGA Reference Clock Constraint",
                    index=adijif.xilinx._ref_clock_constraint_options.index(
                        "Unconstrained"
                    ),
                )
                sys.fpga.ref_clock_constraint = ref_clock_constraint

                # sys.fpga.sys_clk_select = "XCVR_QPLL0"  # Use faster QPLL
                sys_clk_select = st.multiselect(
                    options=adijif.xilinx.sys_clk_selections,
                    label="XCVR System Clock Source Selection",
                    default=adijif.xilinx.sys_clk_selections,
                )
                sys.fpga.sys_clk_select = sys_clk_select

                # Enable all adijif.xilinx._out_clk_selections for selection by default
                out_clk_select = st.multiselect(
                    options=adijif.xilinx._out_clk_selections,
                    label="XCVR Output Clock Selection",
                    default=adijif.xilinx._out_clk_selections,
                )
                sys.fpga.out_clk_select = out_clk_select

                # sys.fpga.force_qpll = 1
                force_qpll_options = ["Auto", "Force QPLL", "Force QPLL1", "Force CPLL"]
                force_qpll_selection = st.selectbox(
                    options=force_qpll_options,
                    label="Transceiver PLL Selection",
                    index=0,
                )
                if force_qpll_selection == "Force QPLL":
                    sys.fpga.force_qpll = True
                elif force_qpll_selection == "Force QPLL1":
                    sys.fpga.force_qpll1 = True
                elif force_qpll_selection == "Force CPLL":
                    sys.fpga.force_cpll = True

        with st.expander("Derived Settings", expanded=True):

            # Table with lane rate and core clock
            lane_rate = sys.converter.bit_clock
            core_clock = (
                sys.converter.bit_clock / 66
                if sys.converter.jesd_class == "jesd204c"
                else sys.converter.bit_clock / 40
            )
            sample_clock = sys.converter.sample_clock
            converter_clock = sys.converter.converter_clock

            # Build pandas DataFrame
            df = pd.DataFrame(
                {
                    "Setting": [
                        "Lane Rate (Gbps)",
                        "Needed Core Clock (MHz)",
                        "Sample Clock (MHz)",
                        "Converter Clock (MHz)",
                    ],
                    "Value": [
                        f"{lane_rate/1e9:.4f}",
                        f"{core_clock/1e6:.4f}",
                        f"{sample_clock/1e6:.4f}",
                        f"{converter_clock/1e6:.4f}",
                    ],
                }
            )
            # Remove index
            df.index = df["Setting"]
            df = df.drop(columns=["Setting"])
            st.table(df)

        with st.expander("System Configuration", expanded=False):
            try:
                cfg = sys.solve()

                st.subheader("Clock Configuration")
                st.write(cfg["clock"])

                st.subheader("Converter Configuration")
                st.write(cfg["converter"])

                st.subheader("FPGA Configuration")
                st.write(cfg["fpga_" + sys.converter.name.upper()])

                st.subheader("Converter JESD Configuration")
                st.write(cfg["jesd_" + sys.converter.name.upper()])

                diagram = sys.draw(cfg)

            except Exception as e:
                diagram = None
                st.error(f"Error solving system configuration: {e}")

        with st.expander("Diagram", expanded=False):
            if diagram:
                st.image(diagram, width="stretch")
            else:
                st.write("No diagram available.")



"""Diagraming functions for different components."""

from __future__ import annotations

import os
from importlib.util import find_spec
from typing import Union

connection_style_types = {
    "data": {"stroke": "blue", "stroke-width": 2},
}


class Node:
    """Node model for diagraming which can have children and connections."""

    def __init__(self, name: str, ntype: str = None, parent: Node = None) -> None:
        """Initialize node with name, type and parent node.

        Args:
            name (str): Name of the node.
            ntype (str): Type of the node.
            parent (Node): Parent node of the node.
        """
        self.name = name
        self.parent = parent
        self.ntype = ntype
        self.children = []
        self.connections = []
        self.shape = "rectangle"
        self.use_unit_conversion_for_rate = True
        self._value = None

    def __repr__(self) -> str:
        """Get string representation of the node.

        Returns:
            str: String representation of the node.
        """
        return f"Node({self.name})"

    @property
    def value(self) -> str:
        """Get value of the node.

        Returns:
            str: Value of the node.
        """
        return f"{self.name} = {self._value}" if self._value else None

    @value.setter
    def value(self, value: Union(int, float, str)) -> None:
        """Set value of the node."""
        self._value = value

    def add_connection(self, connection: dict) -> None:
        """Add connection between this node and another node.

        Args:
            connection (dict): Connection dictionary with keys "from", "to",
                optionally "rate" and "style".
        """
        if "rate" in connection and self.use_unit_conversion_for_rate:
            units = ["Hz", "kHz", "MHz", "GHz"]
            rate = connection["rate"]
            # Convert to unit based on rate scale
            for unit in units:
                if rate < 1000:
                    connection["rate"] = f"{rate:.2f} {unit}"
                    break
                rate /= 1000

        self.connections.append(connection)

    def get_connection(self, from_s: str, to: str) -> dict:
        """Get connection between this node and another node.

        Args:
            from_s (str): Name of the node from which connection originates.
            to (str): Name of the node to which connection goes.

        Returns:
            dict: Connection dictionary with keys "from", "to" and optionally "rate".

        Raises:
            ValueError: If connection not found.
        """
        for conn in self.connections:
            if conn["from"].name == from_s and conn["to"].name == to:
                return conn
        raise ValueError(f"Connection from {from_s} to {to} not found.")

    def remove_connection(self, from_s: str, to: str) -> None:
        """Remove connection between this node and another node.

        Args:
            from_s (str): Name of the node from which connection originates.
            to (str): Name of the node to which connection goes.

        Raises:
            ValueError: If connection not found.
        """
        conn_to_remove = None
        for conn in self.connections:
            if conn["from"].name == from_s and conn["to"].name == to:
                conn_to_remove = conn
                break
        if conn_to_remove:
            self.connections.remove(conn_to_remove)
            return
        raise ValueError(f"Connection from {from_s} to {to} not found.")

    def update_connection(self, from_s: str, to: str, rate: Union(int, float)) -> None:
        """Update connection rate between this node and another node.

        Args:
            from_s (str): Name of the node from which connection originates.
            to (str): Name of the node to which connection goes.
            rate (float): Rate of the connection.

        Raises:
            ValueError: If connection not found.
        """
        for conn in self.connections:
            if conn["from"].name == from_s and conn["to"].name == to:
                units = ["Hz", "kHz", "MHz", "GHz"]
                # Convert to unit based on rate scale
                for unit in units:
                    if rate < 1000:
                        conn["rate"] = f"{rate:.2f} {unit}"
                        return
                    rate /= 1000

        raise ValueError(f"Connection from {from_s} to {to} not found.")

    def add_child(self, child: Node) -> None:
        """Add child node to this node.

        Args:
            child (Node): Child node to add.
        """
        if not isinstance(child, list):
            child = [child]
        for c in child:
            c.parent = self
            self.children.append(c)

    def get_child(self, name: str) -> Node:
        """Get child node by name.

        Args:
            name (str): Name of the child node.

        Returns:
            Node: Child node with the given name.

        Raises:
            ValueError: If child node not found.
        """
        for child in self.children:
            if child.name == name:
                return child
        raise ValueError(f"Child with name {name} not found.")

    def remove_child(self, name: str) -> None:
        """Remove child node by name.

        Args:
            name (str): Name of the child node to remove.
        """
        # Remove connections with the child first
        connections_to_keep = []
        for conn in self.connections:
            if conn["to"].name != name:
                connections_to_keep.append(conn)
        self.connections = connections_to_keep
        connections_to_keep = []
        for conn in self.connections:
            if conn["from"].name != name:
                connections_to_keep.append(conn)
        self.connections = connections_to_keep

        children_to_keep = []
        for child in self.children:
            if child.name != name:
                children_to_keep.append(child)
        self.children = children_to_keep


class Layout:
    """Layout model for diagraming which contains nodes and connections."""

    si = "    "
    use_d2_cli = False
    _write_out_d2_file = False

    def __init__(self, name: str) -> None:
        """Initialize layout with name.

        Args:
            name (str): Name of the layout.
        """
        self.name = name
        self.nodes = []
        self.connections = []
        self.use_unit_conversion_for_rate = True
        self.output_filename = "clocks.d2"
        self.output_image_filename = "clocks.svg"
        self.layout_engine = "elk"
        self.show_rates = True

    def add_node(self, node: Node) -> None:
        """Add node to the layout.

        Args:
            node (Node): Node to add to the layout.
        """
        self.nodes.append(node)

    def remove_node(self, name: str) -> None:
        """Remove node by name.

        Args:
            name (str): Name of the node to remove.
        """
        # Remove connections with the node first
        connections_to_keep = []
        for conn in self.connections:
            if conn["to"].name != name:
                connections_to_keep.append(conn)
        self.connections = connections_to_keep
        connections_to_keep = []
        for conn in self.connections:
            if conn["from"].name != name:
                connections_to_keep.append(conn)
        self.connections = connections_to_keep

        nodes_to_keep = []
        for node in self.nodes:
            if node.name != name:
                nodes_to_keep.append(node)
        self.nodes = nodes_to_keep

    def add_connection(self, connection: dict) -> None:
        """Add connection between two nodes.

        Args:
            connection (dict): Connection dictionary with keys "from", "to"
                and optionally "rate".
        """
        if "rate" in connection and self.use_unit_conversion_for_rate:
            units = ["Hz", "kHz", "MHz", "GHz"]
            rate = connection["rate"]
            # Convert to unit based on rate scale
            for unit in units:
                if rate < 1000:
                    connection["rate"] = f"{rate:.2f} {unit}"
                    break
                rate /= 1000
        self.connections.append(connection)

    def get_connection(
        self, from_s: str = None, to: str = None
    ) -> Union[dict, list[dict]]:
        """Get connection between two nodes.

        Args:
            from_s (str): Name of the node from which connection originates.
            to (str): Name of the node to which connection goes.

        Returns:
            List[dict] or dict: List of or single Connection dictionary with
              keys "from", "to" and optionally "rate".

        Raises:
            ValueError: If connection not found.
        """
        if from_s is None and to is None:
            raise ValueError("Both from and to cannot be None.")
        found = []
        if from_s is None:
            for conn in self.connections:
                if conn["to"].name == to:
                    found.append(conn)
            return found
        if to is None:
            for conn in self.connections:
                if conn["from"].name == from_s:
                    found.append(conn)
            return found
        for conn in self.connections:
            if conn["from"].name == from_s and conn["to"].name == to:
                return conn

    def get_node(self, name: str) -> Node:
        """Get node by name.

        Args:
            name (str): Name of the node.

        Returns:
            Node: Node with the given name.

        Raises:
            ValueError: If node not found.
        """
        for node in self.nodes:
            if node.name == name:
                return node
        raise ValueError(f"Node with name {name} not found.")

    def draw(self) -> str:
        """Draw diagram in d2 language.

        Returns:
            str: Path to the output image file.

        Raises:
            Exception: d2 support not installed.
        """
        diag = "direction: right\n\n"

        def apply_connection_style(diag: str, connection: dict) -> str:
            """Apply style to the connection in the diagram.

            Args:
                diag (str): Current diagram string.
                connection (dict): Connection dictionary with keys "from", "to",
                    and optionally "style".

            Returns:
                str: Updated diagram string with applied style.
            """
            from_p_name = get_parents_names(connection["from"])
            to_p_name = get_parents_names(connection["to"])

            if "type" in connection and connection["type"] in connection_style_types:
                style = connection_style_types[connection["type"]]
                index = 0  # Default index but make sure its not in the diag
                for key in style:
                    while True:
                        con_str = f"({from_p_name}{connection['from'].name} -> "
                        con_str += f"{to_p_name}{connection['to'].name})[{index}]"
                        con_str += f".style.{key}: {style[key]}\n"
                        if con_str not in diag:
                            diag += con_str
                            break
                        index += 1  # Increment index to avoid duplicates
                        # Duplicates can happen if multiple connections
                        # between the same nodes

            index = 0
            if "index" in connection:
                index = connection["index"]
            if "style" in connection:
                for key in connection["style"]:
                    diag += f"({from_p_name}{connection['from'].name} -> "
                    diag += f"{to_p_name}{connection['to'].name})[{index}]"
                    diag += f".style.{key}: {connection['style'][key]}\n"
            return diag

        def get_parents_names(node: Node) -> str:
            """Get names of all parent nodes of the given node.

            Args:
                node (Node): Node for which to get parent names.

            Returns:
                str: Names of all parent nodes of the given node.
            """
            parents = []
            while node.parent:
                parents.append(node.parent.name)
                node = node.parent
            if not parents:
                return ""
            return ".".join(parents[::-1]) + "."

        def draw_subnodes(node: Node, spacing: str = "    ") -> str:
            """Draw subnodes of the given node.

            Args:
                node (Node): Node for which to draw subnodes.
                spacing (str): Spacing for indentation.

            Returns:
                str: Subnodes of the given node.
            """
            diag = " {\n"
            for child in node.children:
                if child.value:
                    diag += spacing + child.name + f": {{tooltip: {child.value} }}"
                else:
                    diag += spacing + child.name
                if child.children:
                    diag += draw_subnodes(child, spacing + "    ")
                else:
                    diag += "\n"
                diag += spacing + child.name + ".shape: " + child.shape + "\n"
            lr = len("    ")
            diag += spacing[:-lr] + "}\n"
            return diag

        # Add all nodes
        for node in self.nodes:
            diag += f"{node.name}"
            if node.children:
                diag += draw_subnodes(node)
            diag += "\n"

        diag += "\n"

        # # Set shapes
        # for node in self.nodes:
        #     parents_string = get_parents_names(node)
        #     diag += f"{parents_string}{node.name}.shape = {node.shape}\n"

        # diag += "\n"

        # Add all connections
        for connection in self.connections:
            from_p_name = get_parents_names(connection["from"])
            to_p_name = get_parents_names(connection["to"])
            if self.show_rates:
                label = f"{connection['rate']}" if "rate" in connection else None
            else:
                label = None
            diag += f"{from_p_name}{connection['from'].name} ->"
            diag += f" {to_p_name}{connection['to'].name}"
            diag += ": " + label if label else ""
            diag += "\n"
            diag = apply_connection_style(diag, connection)

        def draw_nodes_connections(nodes: list[Node], show_rates: bool) -> str:
            """Draw connections between nodes.

            Args:
                nodes (list[Node]): List of nodes to draw connections between.
                show_rates (bool): Whether to show rates in the diagram.

            Returns:
                str: Connections between nodes.
            """
            diag = ""
            for node in nodes:
                for connection in node.connections:
                    from_p_name = get_parents_names(connection["from"])
                    to_p_name = get_parents_names(connection["to"])
                    if show_rates:
                        label = f"{connection['rate']}" if "rate" in connection else ""
                    else:
                        label = ""
                    diag += f"{from_p_name}{connection['from'].name} -> "
                    diag += f"{to_p_name}{connection['to'].name}"
                    diag += ": " + label if label else ""
                    diag += "\n"
                    diag = apply_connection_style(diag, connection)

                if node.children:
                    diag += draw_nodes_connections(node.children, show_rates)

            return diag

        # for node in self.nodes:
        diag += draw_nodes_connections(self.nodes, show_rates=self.show_rates)

        # for node in self.nodes:
        #     for connection in node.connections:
        #         from_p_name = get_parents_names(connection["from"])
        #         to_p_name = get_parents_names(connection["to"])
        #         label = f"{connection['rate']}" if "rate" in connection else ""
        #         diag += f"{from_p_name}{connection['from'].name} -> "
        #         diag += f"{to_p_name}{connection['to'].name}"
        #         diag += ": " + label if label else ""
        #         diag += "\n"

        if self.use_d2_cli:
            with open(self.output_filename, "w") as f:
                f.write(diag)

            cmd = (
                f"d2 --theme=0 --dark-theme=200 -l {self.layout_engine} "
                + f"{self.output_filename} "
            )
            cmd += f"{self.output_image_filename}"
            os.system(cmd)  # noqa: S605
            return self.output_image_filename
        else:
            if self._write_out_d2_file:
                with open(self.output_filename, "w") as f:
                    f.write(diag)
                print(f"Saved to {self.output_filename}")

            # Use bindings
            if find_spec("d2"):
                from d2 import compile
            else:
                raise Exception(
                    "d2 support not installed. Please install package pyd2lang-native"
                )

            out = compile(diag)
            # with open(self.output_image_filename, "w") as f:
            #     f.write(out)

            # return self.output_image_filename
            return out



"""AD9081 Datapath Description Class."""


class ad9081_dp_rx:
    """AD9081 RX Data Path Configuration."""

    cddc_enabled = [True, True, True, True]
    cddc_decimations = [1, 1, 1, 1]
    cddc_decimations_available = [1, 2, 3, 4, 6]
    cddc_nco_frequencies = [0, 0, 0, 0]
    cddc_nco_phases = [0, 0, 0, 0]

    fddc_enabled = [False, False, False, False, False, False, False, False]
    fddc_decimations = [1, 1, 1, 1, 1, 1, 1, 1]
    fddc_decimations_available = [1, 2, 3, 4, 6, 8, 12, 16, 24]
    fddc_nco_frequencies = [0, 0, 0, 0, 0, 0, 0, 0]
    fddc_nco_phases = [0, 0, 0, 0, 0, 0, 0, 0]
    fddc_source = [1, 1, 2, 2, 3, 3, 4, 4]

    __isfrozen = False

    def __init__(self) -> None:
        """Initialize the AD9081 RX datapath."""
        self._freeze()

    def __setattr__(self, key: str, value: any) -> None:
        """Set attribute intercept.

        Only allow setting of attributes that already exist.

        Args:
            key (str): attribute name
            value (any): attribute value

        Raises:
            TypeError: if attribute does not exist
        """
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError("Property %r does not exist" % key)

        if key in ["cddc_decimations", "fddc_decimations"]:
            # Error if decimations are not a list
            if not isinstance(value, list):
                raise TypeError("Decimations must be a list")
            if key == "cddc_decimations":
                if len(value) != 4:
                    raise TypeError("CDDC Decimations must be a list of length 4")
                for v in value:
                    if v not in [1, 2, 3, 4, 6]:
                        raise TypeError("CDDC Decimations must be 1, 2, 3, 4, or 6")
            if key == "fddc_decimations":
                if len(value) != 8:
                    raise TypeError("FDDC Decimations must be a list of length 8")
                for v in value:
                    if v not in [1, 2, 3, 4, 6, 8, 12, 16, 24]:
                        raise TypeError(
                            "FDDC Decimations must be 1, 2, 4, 6, 8, 12, 16, or 24"
                        )
        object.__setattr__(self, key, value)

    def _freeze(self) -> None:
        self.__isfrozen = True

    def get_config(self) -> dict:
        """Get the datapath configuration for the AD9081 RX.

        Returns:
            dict: datapath configuration
        """
        datapath = {}
        datapath["cddc"] = {}
        datapath["cddc"]["enabled"] = self.cddc_enabled
        datapath["cddc"]["decimations"] = self.cddc_decimations
        datapath["cddc"]["nco_frequencies"] = self.cddc_nco_frequencies
        datapath["cddc"]["nco_phases"] = self.cddc_nco_phases

        datapath["fddc"] = {}
        datapath["fddc"]["enabled"] = self.fddc_enabled
        datapath["fddc"]["decimations"] = self.fddc_decimations
        datapath["fddc"]["nco_frequencies"] = self.fddc_nco_frequencies
        datapath["fddc"]["nco_phases"] = self.fddc_nco_phases
        datapath["fddc"]["source"] = self.fddc_source

        return datapath

    @property
    def decimation_overall(self) -> int:
        """Minimum Overall Decimation factor.

        Raises:
            Exception: No FDDC or CDDC enabled
            Exception: Enabled FDDC's source CDDC not enabled

        Returns:
            int: minimum overall decimation factor
        """
        if (not any(self.fddc_enabled)) and (not any(self.cddc_enabled)):
            raise Exception("No FDDCs or CDDCs enabled")

        min_dec = -1
        if any(self.fddc_enabled):
            for i, fdec in enumerate(self.fddc_decimations):
                if self.fddc_enabled[i]:
                    cddc = self.fddc_source[i] - 1
                    if not self.cddc_enabled:
                        raise Exception(f"Source CDDC {cddc} not enabled for FDDC {i}")
                    cdec = self.cddc_decimations[cddc]
                    if (cdec * fdec < min_dec) or min_dec == -1:
                        min_dec = cdec * fdec
        else:
            for i, cdec in enumerate(self.cddc_decimations):
                if self.cddc_enabled[i] and (min_dec == -1 or cdec < min_dec):
                    min_dec = cdec

        return min_dec


class ad9081_dp_tx:
    """AD9081 TX Data Path Configuration."""

    cduc_enabled = [True, True, True, True]
    cduc_interpolation = 1
    cduc_interpolations_available = [1, 2, 4, 6, 8, 12]
    cduc_nco_frequencies = [0, 0, 0, 0]
    cduc_nco_phases = [0, 0, 0, 0]

    fduc_enabled = [False, False, False, False, False, False, False, False]
    fduc_interpolation = 1
    fduc_interpolations_available = [1, 2, 3, 4, 6, 8]
    fduc_nco_frequencies = [0, 0, 0, 0, 0, 0, 0, 0]
    fduc_nco_phases = [0, 0, 0, 0, 0, 0, 0, 0]

    cduc_sources = [[1], [1], [3], [3]]

    __isfrozen = False

    def __init__(self) -> None:
        """Initialize the AD9081 TX datapath."""
        self._freeze()

    def __setattr__(self, key: str, value: any) -> None:
        """Set attribute intercept.

        Only allow setting of attributes that already exist.

        Args:
            key (str): attribute name
            value (any): attribute value

        Raises:
            TypeError: if attribute does not exist
        """
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError("Property %r does not exist" % key)

        if key in ["cduc_interpolation", "fduc_interpolation"]:
            # Error if interpolation is a list
            if isinstance(value, list):
                raise TypeError("Interpolation must be an integer, not a list")
            if key == "cduc_interpolation" and value not in [1, 2, 4, 6, 8, 12]:
                raise TypeError("CDUC Interpolation must be 1, 2, 4, 6, 8, or 12")
            if key == "fduc_interpolation" and value not in [1, 2, 3, 4, 6, 8]:
                raise TypeError("FDUC Interpolation must be 1, 2, 3, 4, 6, or 8")
        object.__setattr__(self, key, value)

    def _freeze(self) -> None:
        self.__isfrozen = True

    def get_config(self) -> dict:
        """Get the datapath configuration for the AD9081 TX.

        Returns:
            dict: Datapath configuration
        """
        datapath = {}
        datapath["cduc"] = {}
        datapath["cduc"]["enabled"] = self.cduc_enabled
        datapath["cduc"]["interpolation"] = self.cduc_interpolation
        datapath["cduc"]["nco_frequencies"] = self.cduc_nco_frequencies
        datapath["cduc"]["nco_phases"] = self.cduc_nco_phases
        datapath["cduc"]["sources"] = self.cduc_sources

        datapath["fduc"] = {}
        datapath["fduc"]["enabled"] = self.fduc_enabled
        datapath["fduc"]["interpolation"] = self.fduc_interpolation
        datapath["fduc"]["nco_frequencies"] = self.fduc_nco_frequencies
        datapath["fduc"]["nco_phases"] = self.fduc_nco_phases

        return datapath

    @property
    def interpolation_overall(self) -> int:
        """Minimum Overall Interpolation factor.

        Raises:
            Exception: No FDDC or CDDC enabled
            Exception: Enabled FDDC's source CDDC not enabled

        Returns:
            int: minimum overall interpolation factor
        """
        if (not any(self.fduc_enabled)) and (not any(self.cduc_enabled)):
            raise Exception("No FDUCs or CDUCs enabled")

        if any(self.fduc_enabled):
            return self.fduc_interpolation * self.cduc_interpolation
        else:
            return self.cduc_interpolation



"""ADF4371 Microwave Wideband Synthesizer with Integrated VCO model."""

from typing import Dict, List, Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.plls.pll import pll
from adijif.solvers import CpoExpr, GK_Intermediate, integer_var, tround


class adf4371(pll):
    """ADF4371 PLL model.

    This model currently supports all divider configurations

    https://www.analog.com/media/en/technical-documentation/data-sheets/adf4371.pdf
    """

    name = "adf4371"

    input_freq_max = int(600e6)
    input_freq_min = int(10e6)

    pfd_freq_max_frac = int(160e6)
    pfd_freq_max_int = int(250e6)

    vco_freq_max = int(8e9)
    vco_freq_min = int(4e9)

    _d = [0, 1]
    d_available = [0, 1]

    @property
    def d(self) -> Union[int, List[int]]:
        """REF-in doubler.

        Valid values are 1,2

        Returns:
            int: Current allowable setting
        """
        return self._d

    @d.setter
    def d(self, value: Union[int, List[int]]) -> None:
        """REF-in double.

        Valid values are 1,2

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.d_available, "d")
        self._d = value

    _r = [*range(1, 32 + 1)]
    r_available = [*range(1, 32 + 1)]

    @property
    def r(self) -> Union[int, List[int]]:
        """Reference divider.

        Valid values are 0->(2^5-1)

        Returns:
            int: Current allowable setting
        """
        return self._r

    @r.setter
    def r(self, value: Union[int, List[int]]) -> None:
        """Reference divider.

        Valid values are 1->32

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.r_available, "r")
        self._r = value

    _t = [0, 1]
    t_available = [0, 1]

    @property
    def t(self) -> Union[int, List[int]]:
        """Reference divide by 2.

        Valid values are 0,1

        Returns:
            int: Current allowable setting
        """
        return self._t

    @t.setter
    def t(self, value: Union[int, List[int]]) -> None:
        """Reference divide by 2.

        Valid values are 0,1

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.t_available, "t")
        self._t = value

    _rf_div = [1, 2, 4, 8, 16, 32, 64]
    rf_div_available = [1, 2, 4, 8, 16, 32, 64]

    @property
    def rf_div(self) -> Union[int, List[int]]:
        """Output RF divider.

        Valid dividers are 1,2,3,4,5,6..32->(even)->4096

        Returns:
            int: Current allowable dividers
        """
        return self._rf_div

    @rf_div.setter
    def rf_div(self, value: Union[int, List[int]]) -> None:
        """Output RF divider.

        Valid dividers are 1,2,3,4,5,6..32->(even)->4096

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.rf_div_available, "rf_div")
        self._rf_div = value

    _mode = ["integer", "fractional"]
    mode_available = ["integer", "fractional"]

    @property
    def mode(self) -> Union[str, List[str]]:
        """Set operational mode.

        Options are: fractional, integer or [fractional, integer]

        Returns:
            str: Current allowable modes
        """
        return self._mode

    @mode.setter
    def mode(self, value: Union[str, List[str]]) -> None:
        """Set operational mode.

        Options are: fractional, integer or [fractional, integer]

        Args:
            value (str, list[str]): Allowable values for mode

        """
        self._check_in_range(value, self.mode_available, "mode")
        self._mode = value

    # These are too large for user to set
    _int_4d5_min_max = [20, 32767]
    _int_8d9_min_max = [64, 65535]
    _int_frac_4d5_min_max = [23, 32767]
    _int_frac_8d9_min_max = [75, 65535]

    _frac1_min_max = [0, 33554431]
    _frac2_min_max = [0, 16383]
    _MOD1 = 2**25
    _MOD2 = [*range(2, 16383 + 1)]

    _prescaler = ["4/5", "8/9"]

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        config: Dict = {
            "d": self._get_val(self.config["d"]),
            "r": self._get_val(self.config["r"]),
            "t": self._get_val(self.config["t"]),
            "frac1": self._get_val(self.config["frac1"]),
            "frac2": self._get_val(self.config["frac2"]),
            "MOD2": self._get_val(self.config["MOD2"]),
            "int": self._get_val(self.config["int"]),
            "rf_div": self._get_val(self.config["rf_div"]),
        }

        if self._get_val(self.config["fact_0_int_1"]) == 1:
            config["mode"] = "fractional"
        else:
            config["mode"] = "integer"

        if self._get_val(self.config["prescaler_4/5_0_8/9_1"]) == 1:
            config["prescaler"] = "8/9"
        else:
            config["prescaler"] = "4/5"

        vco = self.solution.get_kpis()["vco"]
        config["rf_out_frequency"] = vco / config["rf_div"]
        config["rf_out_frequency"] = tround(config["rf_out_frequency"])

        return config

    def _setup_solver_constraints(
        self, input_ref: Union[int, float, CpoExpr, GK_Intermediate]
    ) -> None:
        """Apply constraints to solver model.

        Args:
            input_ref (int, float, CpoExpr, GK_Intermediate): Input reference
                frequency in hertz. Can also be range or arb_source type.

        Raises:
            NotImplementedError: If solver is not CPLEX
        """
        self.config = {}

        # Handle range type (returns dict with "range" key)
        # arb_source and direct expressions are already supported
        expr_types = tuple(filter(None, [int, float, CpoExpr, GK_Intermediate]))
        if not isinstance(input_ref, expr_types):
            input_ref_result = input_ref(self.model)  # type: ignore
            if isinstance(input_ref_result, dict):
                # range type
                self.config["input_ref_set"] = input_ref_result
                input_ref = self.config["input_ref_set"]["range"]
            else:
                # arb_source type (returns expression directly)
                input_ref = input_ref_result
        self.input_ref = input_ref

        # PFD
        self.config["d"] = self._convert_input(self.d, name="d")
        self.config["r"] = self._convert_input(self.r, name="r")
        self.config["t"] = self._convert_input(self.t, name="t")

        self.config["f_pfd"] = self._add_intermediate(
            input_ref
            * ((1 + self.config["d"]) / (self.config["r"] * (1 + self.config["t"])))
        )

        # Configure fractional mode or integer mode constraints
        if self._mode == "fractional":
            self.config["fact_0_int_1"] = self._convert_input(0, "fact_0_int_1")
        elif self._mode == "integer":
            self.config["fact_0_int_1"] = self._convert_input(1, "fact_0_int_1")
        else:
            self.config["fact_0_int_1"] = self._convert_input([0, 1], "fact_0_int_1")

        self.config["pfd_max_freq"] = self._add_intermediate(
            (1 - self.config["fact_0_int_1"]) * self.pfd_freq_max_frac
            + self.config["fact_0_int_1"] * self.pfd_freq_max_int,
        )

        # VCO + output
        self.config["MOD2"] = self._convert_input(self._MOD2, name="MOD2")

        # Configure INT setting based on prescalers
        if self.solver == "CPLEX":
            self.config["frac1"] = integer_var(
                min=self._frac1_min_max[0],
                max=self._frac1_min_max[1],
                name="frac1",
            )
            self.config["frac2"] = integer_var(
                min=self._frac2_min_max[0],
                max=self._frac2_min_max[1],
                name="frac2",
            )

            if self._prescaler == "4/5":
                self.config["prescaler_4/5_0_8/9_1"] = self._convert_input(
                    0, "prescaler"
                )
            elif self._prescaler == "8/9":
                self.config["prescaler_4/5_0_8/9_1"] = self._convert_input(
                    1, "prescaler"
                )
            else:
                self.config["prescaler_4/5_0_8/9_1"] = self._convert_input(
                    [0, 1], "prescaler"
                )

            self.config["int_min"] = self._add_intermediate(
                self.config["fact_0_int_1"]
                * (
                    self.config["prescaler_4/5_0_8/9_1"] * self._int_8d9_min_max[0]
                    + (1 - self.config["prescaler_4/5_0_8/9_1"])
                    * self._int_4d5_min_max[0]
                )
                + (1 - self.config["fact_0_int_1"])
                * (
                    self.config["prescaler_4/5_0_8/9_1"] * self._int_frac_8d9_min_max[0]
                    + (1 - self.config["prescaler_4/5_0_8/9_1"])
                    * self._int_frac_4d5_min_max[0]
                )
            )

            self.config["int_max"] = self._add_intermediate(
                self.config["fact_0_int_1"]
                * (
                    self.config["prescaler_4/5_0_8/9_1"] * self._int_8d9_min_max[1]
                    + (1 - self.config["prescaler_4/5_0_8/9_1"])
                    * self._int_4d5_min_max[1]
                )
                + (1 - self.config["fact_0_int_1"])
                * (
                    self.config["prescaler_4/5_0_8/9_1"] * self._int_frac_8d9_min_max[1]
                    + (1 - self.config["prescaler_4/5_0_8/9_1"])
                    * self._int_frac_4d5_min_max[1]
                )
            )

            min_o = min(
                [
                    self._int_8d9_min_max[0],
                    self._int_4d5_min_max[0],
                    self._int_frac_8d9_min_max[0],
                    self._int_frac_4d5_min_max[0],
                ]
            )
            max_o = max(
                [
                    self._int_8d9_min_max[1],
                    self._int_4d5_min_max[1],
                    self._int_frac_8d9_min_max[1],
                    self._int_frac_4d5_min_max[1],
                ]
            )
            self.config["int"] = integer_var(min=min_o, max=max_o, name="int")
        else:
            raise NotImplementedError("Only CPLEX solver is implemented")

        self.config["vco"] = self._add_intermediate(
            (
                self.config["int"]
                + (self.config["frac1"] + self.config["frac2"] / self.config["MOD2"])
                / self._MOD1
            )
            * self.config["f_pfd"]
        )
        self.model.add_kpi(
            (
                self.config["int"]
                + (self.config["frac1"] + self.config["frac2"] / self.config["MOD2"])
                / self._MOD1
            )
            * self.config["f_pfd"],
            "vco",
        )

        self._add_equation(
            [
                input_ref <= self.input_freq_max,
                input_ref >= self.input_freq_min,
                self.config["f_pfd"] <= self.config["pfd_max_freq"],
                self.config["vco"] <= self.vco_freq_max,
                self.config["vco"] >= self.vco_freq_min,
                self.config["int"] <= self.config["int_max"],
                self.config["int"] >= self.config["int_min"],
            ]
        )

    def _setup(self, input_ref: int) -> None:
        if isinstance(input_ref, (float, int)):
            assert (
                self.input_freq_max >= input_ref >= self.input_freq_min
            ), "Input frequency out of range"

        # Setup clock chip internal constraints
        self._setup_solver_constraints(input_ref)

    def _get_clock_constraint(
        self, clk_name: str
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  Reference clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference
        """
        self._clk_names = ["clk_name"]

        self.config["rf_div"] = self._convert_input(self.rf_div, "rf_div")

        return self.config["vco"] / self.config["rf_div"]

    def set_requested_clocks(
        self, ref_in: Union[int, float, CpoExpr, GK_Intermediate], out_freq: int
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            ref_in (int, float, CpoExpr, GK_Intermediate): Reference frequency in hertz
            out_freq (int): list of required clocks to be output

        """
        self._setup(ref_in)
        self._clk_names = ["rf_out"]

        self.config["rf_div"] = self._convert_input(self.rf_div, "rf_div")

        self._add_equation([self.config["rf_div"] * out_freq == self.config["vco"]])



"""Xilinx Common PLL class."""

from typing import Optional, Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from ...common import core
from ...gekko_trans import gekko_translation
from ...solvers import CpoModel


class XilinxPLL(core, gekko_translation):
    """Xilinx Common PLL class."""

    plls = None
    parent = None
    _model = None  # Hold internal model when used standalone
    _solution = None  # Hold internal solution when used standalone

    def __init__(
        self,
        parent=None,  # noqa: ANN001
        speed_grade: Optional[str] = "-2",
        transceiver_type: Optional[str] = "GTXE2",
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> None:
        """Initalize 7 series transceiver PLLs.

        Args:
            parent (system or converter, optional): Parent object. Defaults to None.
            speed_grade (str, optional): Speed grade. Defaults to "-2".
            transceiver_type (str, optional): Transceiver type. Defaults to "GTXE2".
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            Exception: If Gekko solver is used
        """
        self.transceiver_type = transceiver_type
        self.speed_grade = speed_grade
        super().__init__(*args, **kwargs)
        self.parent = parent
        if parent:
            self._model = parent.model
            self._solution = parent.solution
            self.solver = parent.solver
        self.add_plls()
        if self.solver == "gekko":
            raise Exception("Gekko solver not supported for Xilinx PLLs")

    @property
    def model(self) -> CpoModel:
        """Internal system model for solver.

        Returns:
            CpoModel: Internal system model for solver
        """
        if self.parent:
            return self.parent.model
        return self._model

    @model.setter
    def model(self, val: CpoModel) -> None:
        """Set internal system model for solver.

        Args:
            val (CpoModel): Internal system model for solver

        Raises:
            Exception: If parent model is used
        """
        if self.parent:
            raise Exception("Cannot set model when parent model is used")
        self._model = val

    @property
    def solution(self) -> CpoSolveResult:
        """Solution object from solver."""
        if self.parent:
            return self.parent.solution
        return self._solution

    @solution.setter
    def solution(self, val: CpoSolveResult) -> None:
        """Set solution object from solver.

        Args:
            val (CpoSolveResult): Solution object from solver

        Raises:
            Exception: If parent model is used
        """
        if self.parent:
            raise Exception("Cannot set solution when parent model is used")
        self._solution = val

    @property
    def transceiver_type(self) -> str:
        """Transceiver type.

        Returns:
            str: Transceiver type
        """
        return self._transceiver_type

    @transceiver_type.setter
    def transceiver_type(self, val: str) -> None:
        """Set transceiver type.

        Args:
            val (str): Transceiver type
        """
        self._check_in_range(val, self.transceiver_types_available, "transceiver_type")
        self._transceiver_type = val

    _speed_grade = -2

    @property
    def speed_grade(self) -> str:
        """speed_grade for transceiver.

        Returns:
            str: Speed grade
        """
        if self.parent:
            return self.parent.speed_grade
        return self._speed_grade

    @speed_grade.setter
    def speed_grade(self, val: str) -> None:
        """Set speed grade for transceiver.

        Args:
            val (str): Speed grade

        Raises:
            Exception: If parent model is used
        """
        if self.parent:
            raise Exception("Cannot set speed_grade when parent model is used")
        self._speed_grade = val

    def _solve_gekko(self) -> bool:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            bool: Always False
        """
        self.model.options.SOLVER = 1  # APOPT solver
        self.model.solver_options = [
            "minlp_maximum_iterations 1000",  # minlp iterations with integer solution
            "minlp_max_iter_with_int_sol 100",  # treat minlp as nlp
            "minlp_as_nlp 0",  # nlp sub-problem max iterations
            "nlp_maximum_iterations 500",  # 1 = depth first, 2 = breadth first
            "minlp_branch_method 1",  # maximum deviation from whole number
            "minlp_integer_tol 0",  # covergence tolerance (MUST BE 0 TFC)
            "minlp_gap_tol 0.1",
        ]

        self.model.solve(disp=False)
        self.model.cleanup()
        return False

    # def _add_objective(self, sysrefs: List) -> None:
    #     pass

    def _solve_cplex(self) -> CpoSolveResult:
        self.solution = self.model.solve(LogVerbosity="Normal")
        if self.solution.solve_status not in ["Feasible", "Optimal"]:
            raise Exception("Solution Not Found")
        return self.solution

    def solve(self) -> Union[None, CpoSolveResult]:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            [None,CpoSolveResult]: When cplex solver is used CpoSolveResult is returned

        Raises:
            Exception: If solver is not valid

        """
        if self.solver == "gekko":
            return self._solve_gekko()
        elif self.solver == "CPLEX":
            return self._solve_cplex()
        else:
            raise Exception(f"Unknown solver {self.solver}")


class PLLCommon(gekko_translation):
    """Common PLL class for Xilinx and Intel PLLs."""

    def __init__(self, parent_transceiver: CpoModel) -> None:
        """Initialize PLL common class.

        Args:
            parent_transceiver (CpoModel): Parent transceiver object
        """
        self.parent = parent_transceiver

    @property
    def model(self) -> CpoModel:
        """Internal system model for solver.

        Returns:
            CpoModel: Internal system model for solver
        """
        return self.parent.model

    @property
    def solver(self) -> str:
        """Solver type.

        Returns:
            str: Solver type
        """
        return self.parent.solver

    @property
    def solution(self) -> CpoSolveResult:
        """Solution object from solver.

        Returns:
            CpoSolveResult: Solution object from solver
        """
        return self.parent.solution



"""ADF4030 10-Channel Precision Synchronizer."""

from typing import Dict, List, Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.clocks.clock import clock as clockc
from adijif.plls.pll import pll
from adijif.solvers import CpoExpr, GK_Intermediate


class adf4030(pll):
    """ADF4030 PLL model.

    This model currently supports all divider configurations

    https://www.analog.com/media/en/technical-documentation/data-sheets/adf4030.pdf
    """

    name = "adf4030"

    input_freq_min = int(10e6)
    input_freq_max = int(250e6)

    pfd_freq_min = int(10e6)
    pfd_freq_max = int(20e6)

    vco_freq_min = int(2.5e9 * 0.95)
    vco_freq_max = int(2.5e9 * 1.05)

    _r = [*range(1, 31 + 1)]
    r_available = [*range(1, 31 + 1)]

    @property
    def r(self) -> Union[int, List[int]]:
        """Reference divider.

        Valid values are 1->31

        Returns:
            int: Current allowable setting
        """
        return self._r

    @r.setter
    def r(self, value: Union[int, List[int]]) -> None:
        """Reference divider.

        Valid values are 1->31

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.r_available, "r")
        self._r = value

    _n = [*range(8, 255 + 1)]
    n_available = [*range(8, 255 + 1)]

    @property
    def n(self) -> Union[int, List[int]]:
        """Feedback divider.

        Valid values are 8->255

        Returns:
            int: Current allowable setting
        """
        return self._n

    @n.setter
    def n(self, value: Union[int, List[int]]) -> None:
        """Feedback divider.

        Valid values are 8->255

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.n_available, "n")
        self._n = value

    _o = [*range(10, 4095 + 1)]
    o_available = [*range(10, 4095 + 1)]

    @property
    def o(self) -> Union[int, List[int]]:
        """Output divider.

        Valid values are 10->4095

        Returns:
            int: Current allowable setting
        """
        return self._o

    @o.setter
    def o(self, value: Union[int, List[int]]) -> None:
        """Output divider.

        Valid values are 10->4095

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.o_available, "o")
        self._o = value

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        out_dividers = [self._get_val(x) for x in self.config["out_dividers"]]

        config: Dict = {
            "r": self._get_val(self.config["r"]),
            "n": self._get_val(self.config["n"]),
            "out_dividers": out_dividers,
        }

        vco = self.solution.get_kpis()["vco_adf4030"]
        config["vco"] = vco

        # Outputs
        output_config = {}
        for i, clk in enumerate(self._clk_names):
            o_val = out_dividers[i]
            output_config[clk] = {
                "rate": vco / o_val,
                "divider": o_val,
            }

        config["output_clocks"] = output_config

        return config

    def _setup_solver_constraints(
        self, input_ref: Union[int, float, CpoExpr, GK_Intermediate]
    ) -> None:
        """Apply constraints to solver model.

        Args:
            input_ref (int, float, CpoExpr, GK_Intermediate): Input reference
                frequency in hertz. Can also be range or arb_source type.
        """
        self.config = {}

        # Handle range type (returns dict with "range" key)
        # arb_source and direct expressions are already supported
        expr_types = tuple(filter(None, [int, float, CpoExpr, GK_Intermediate]))
        if not isinstance(input_ref, expr_types):
            input_ref_result = input_ref(self.model)  # type: ignore
            if isinstance(input_ref_result, dict):
                # range type
                self.config["input_ref_set"] = input_ref_result
                input_ref = self.config["input_ref_set"]["range"]
            else:
                # arb_source type (returns expression directly)
                input_ref = input_ref_result
        self.input_ref = input_ref

        # PFD
        self.config["r"] = self._convert_input(self.r, name="r_adf4030")
        self.config["n"] = self._convert_input(self.n, name="n_adf4030")
        self.config["o"] = self._convert_input(self.o, name="o_adf4030")

        self.config["vco"] = self._add_intermediate(
            input_ref * self.config["n"] / self.config["r"]
        )
        self.model.add_kpi(
            input_ref * self.config["n"] / self.config["r"],
            "vco_adf4030",
        )

        self._add_equation(
            [
                input_ref <= self.input_freq_max,
                input_ref >= self.input_freq_min,
                self.config["vco"] <= self.vco_freq_max,
                self.config["vco"] >= self.vco_freq_min,
                input_ref <= self.pfd_freq_max * self.config["r"],
                input_ref >= self.pfd_freq_min * self.config["r"],
            ]
        )

    def _setup(self, input_ref: Union[int, clockc]) -> None:
        # For integer/float values, validate frequency range
        if isinstance(input_ref, (float, int)):
            assert (
                self.input_freq_max >= input_ref >= self.input_freq_min
            ), "Input frequency out of range"

        # Setup clock chip internal constraints (this converts input_ref to solver var
        # and adds constraints for range/arb_source types)
        self._setup_solver_constraints(input_ref)

        self._clk_names = []  # List of clock names to be generated
        self.config["out_dividers"] = []

    def _get_clock_constraint(
        self, clk_name: str
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  Reference clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference
        """
        od = self._convert_input(self._o, f"o_div_{clk_name}_adf4030")

        # Update diagram to include new divider
        # d_n = len(self.config["out_dividers"])
        # self._update_diagram({f"o{d_n}": od})

        self._clk_names.append(clk_name)

        self.config["out_dividers"].append(od)
        return self.config["vco"] / od

    def set_requested_clocks(
        self,
        ref_in: Union[int, float, CpoExpr, GK_Intermediate],
        out_freq: Union[int, List[int]],
        clk_names: List[str],
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            ref_in (int, float, CpoExpr, GK_Intermediate): Reference frequency in hertz
            out_freq (int): list of required clocks to be output
            clk_names (List[str]): list of clock names

        Raises:
            Exception: If out_freq and clk_names are not the same length
        """
        if len(out_freq) != len(clk_names):
            raise Exception("out_freq and clk_names must be the same length")
        self._setup(ref_in)
        self._clk_names = clk_names

        for i, clk in enumerate(clk_names):
            o_div_name = f"o_div_{clk}_adf4030"
            self.config[o_div_name] = self._convert_input(self.o, o_div_name)
            self.config["out_dividers"].append(self.config[o_div_name])

            self._add_equation(
                self.config[o_div_name] * out_freq[i] == self.config["vco"],
            )



# flake8: noqa

import pytest

import adijif

from .common import skip_solver


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
@pytest.mark.parametrize("converter", ["adrv9009_rx", "adrv9009_tx"])
def test_adrv9009_rxtx_ad9528_solver_compact(solver, converter):
    skip_solver(solver)
    if solver == "gekko":
        pytest.xfail("gekko currently unsupported")

    vcxo = 122.88e6

    sys = adijif.system(converter, "ad9528", "xilinx", vcxo, solver=solver)

    # Get Converter clocking requirements
    sys.converter.sample_clock = 122.88e6

    if converter == "adrv9009_rx":
        sys.converter.decimation = 4
    else:
        sys.converter.interpolation = 4

    sys.converter.L = 2
    sys.converter.M = 4
    sys.converter.N = 16
    sys.converter.Np = 16

    sys.converter.K = 32
    sys.converter.F = 4
    assert sys.converter.S == 1
    # sys.Debug_Solver = True

    assert 9830.4e6 / 2 == sys.converter.bit_clock
    assert sys.converter.multiframe_clock == 7.68e6 / 2  # LMFC
    assert sys.converter.device_clock == 9830.4e6 / 2 / 40

    # Set FPGA config
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.out_clk_select = "XCVR_REFCLK"
    sys.fpga.force_qpll = True

    # Set clock chip
    sys.clock.d = [*range(1, 257)]  # Limit output dividers

    cfg = sys.solve()
    # print(cfg)
    from pprint import pprint

    pprint(cfg)

    ref = {
        "gekko": {"clock": {"r1": 1, "n2": 8, "m1": 4, "out_dividers": [1, 4, 8, 256]}},
        "CPLEX": {"clock": {"r1": 1, "n2": 6, "m1": 5, "out_dividers": [6, 192]}},
    }

    assert cfg["clock"]["r1"] == ref[solver]["clock"]["r1"]
    assert cfg["clock"]["n2"] == ref[solver]["clock"]["n2"]
    assert cfg["clock"]["m1"] == ref[solver]["clock"]["m1"]
    assert (
        cfg["clock"]["output_clocks"][f"{sys.fpga.name}_{converter.upper()}_ref_clk"][
            "rate"
        ]
        == 122880000.0
    )  # 98304000
    for div in cfg["clock"]["out_dividers"]:
        assert div in ref[solver]["clock"]["out_dividers"]


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_adrv9009_ad9528_solver_compact(solver):
    skip_solver(solver)
    vcxo = 122.88e6
    sys = adijif.system("adrv9009", "ad9528", "xilinx", vcxo, solver=solver)

    # Rx
    sys.converter.adc.sample_clock = 122.88e6
    sys.converter.adc.decimation = 4
    sys.converter.adc.L = 2
    sys.converter.adc.M = 4
    sys.converter.adc.N = 16
    sys.converter.adc.Np = 16

    sys.converter.adc.K = 32
    sys.converter.adc.F = 4
    assert sys.converter.adc.S == 1

    assert 9830.4e6 / 2 == sys.converter.adc.bit_clock
    assert sys.converter.adc.multiframe_clock == 7.68e6 / 2  # LMFC
    assert sys.converter.adc.device_clock == 9830.4e6 / 2 / 40

    # Tx
    sys.converter.dac.sample_clock = 122.88e6
    sys.converter.dac.interpolation = 4
    sys.converter.dac.L = 2
    sys.converter.dac.M = 4
    sys.converter.dac.N = 16
    sys.converter.dac.Np = 16

    sys.converter.dac.K = 32
    sys.converter.dac.F = 4
    assert sys.converter.dac.S == 1

    assert 9830.4e6 / 2 == sys.converter.dac.bit_clock
    assert sys.converter.dac.multiframe_clock == 7.68e6 / 2  # LMFC
    assert sys.converter.dac.device_clock == 9830.4e6 / 2 / 40

    # Set FPGA config
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.out_clk_select = "XCVR_REFCLK"
    sys.fpga.force_qpll = True

    # Set clock chip
    sys.clock.d = [*range(1, 257)]  # Limit output dividers

    if solver == "gekko":
        with pytest.raises(AssertionError):
            cfg = sys.solve()
        pytest.xfail("gekko currently unsupported")

    cfg = sys.solve()
    print(cfg)

    ref = {
        "gekko": {"clock": {"r1": 1, "n2": 8, "m1": 4, "out_dividers": [1, 8, 256]}},
        "CPLEX": {
            "clock": {
                "r1": 1,
                "n2": 6,
                "m1": 5,
                "out_dividers": [1, 6, 192],
            }
        },
    }

    assert cfg["clock"]["r1"] == ref[solver]["clock"]["r1"]
    assert cfg["clock"]["n2"] == ref[solver]["clock"]["n2"]
    assert cfg["clock"]["m1"] == ref[solver]["clock"]["m1"]

    output_clocks = cfg["clock"]["output_clocks"]
    assert output_clocks["zc706_adc_ref_clk"]["rate"] == 122880000.0
    assert output_clocks["zc706_dac_ref_clk"]["rate"] == 122880000.0

    for div in cfg["clock"]["out_dividers"]:
        assert div in ref[solver]["clock"]["out_dividers"]


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_adrv9009_ad9528_quick_config(solver):
    skip_solver(solver)
    vcxo = 122.88e6

    sys = adijif.system("adrv9009", "ad9528", "xilinx", vcxo, solver=solver)
    sys.converter.adc.sample_clock = 122.88e6
    sys.converter.dac.sample_clock = 122.88e6

    sys.converter.adc.decimation = 4
    sys.converter.dac.interpolation = 4

    mode_rx = "17"
    mode_tx = "6"
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")

    assert sys.converter.adc.L == 2
    assert sys.converter.adc.M == 4
    assert sys.converter.adc.N == 16
    assert sys.converter.adc.Np == 16

    assert sys.converter.dac.L == 2
    assert sys.converter.dac.M == 4
    assert sys.converter.dac.N == 16
    assert sys.converter.dac.Np == 16

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    sys.fpga.setup_by_dev_kit_name("zc706")

    if solver == "gekko":
        with pytest.raises(AssertionError):
            cfg = sys.solve()
        pytest.xfail("gekko currently unsupported")

    cfg = sys.solve()



"""System level interface for manage clocks across all devices."""

import os
import shutil  # noqa: F401
from typing import Dict, List, Tuple, Union

import numpy as np

import adijif  # noqa: F401
import adijif.solvers as solvers
from adijif.clocks.clock import clock as clockc
from adijif.converters.converter import converter as convc
from adijif.plls.pll import pll as pllc
from adijif.sys.s_plls import SystemPLL
from adijif.system_draw import system_draw as system_draw
from adijif.types import arb_source as arb_sourcec
from adijif.types import range as rangec


class system(SystemPLL, system_draw):
    """System Manager Class.

    Manage requirements from all system components and feed into clock rate
    generation algorithms

    All converters shared a common sysref. This requires that all
        converters have the same multiframe clock (LMFC)
    """

    use_common_sysref = False

    enable_converter_clocks = True
    enable_fpga_clocks = True

    Debug_Solver = False
    solver = "CPLEX"
    solution = None

    _plls = []

    @property
    def plls(self) -> List[pllc]:
        """External PLLs used to drive converters.

        Returns:
            List: List of PLL objects
        """
        return self._plls

    def add_pll_inline(self, pll_name: str, clk: clockc, cnv: convc) -> None:
        """Add External PLL to system between clock chip and converter.

        Args:
            pll_name (str): Name of PLL class
            clk (clockc): Clock chip reference
            cnv (convc): Converter reference
        """
        pll = eval(f"adijif.{pll_name}(self.model,solver=self.solver)")  # noqa: S307
        self._plls.append(pll)
        pll._connected_to_output = cnv.name

    def _model_reset(self) -> None:
        if self.solver == "gekko":
            if not solvers.gekko_solver:
                raise Exception("GEKKO Solver not installed")
            model = solvers.GEKKO(remote=False)
        elif self.solver == "CPLEX":
            if not solvers.cplex_solver:
                raise Exception("CPLEX Solver not installed")
            model = solvers.CpoModel()
        else:
            raise Exception(f"Unknown solver {self.solver}")

        self.model = model
        self.clock.model = model
        self.clock.config = {}
        self.fpga.model = model
        self.fpga.config = {}
        if isinstance(self.converter, list):
            for conv in self.converter:
                conv.model = model
                conv.config = {}
        else:
            self.converter.config = {}
            self.converter.model = model

        self._plls = []

    def __init__(
        self,
        conv: Union[str, List[str]],
        clk: str,
        fpga: str,
        vcxo: Union[int, float, rangec, arb_sourcec],
        solver: str = None,
    ) -> None:
        """Initialize system interface and manage all clocking aspects.

        Args:
            conv (str): Name of converter class
            clk (str): Name of Clock chip class
            fpga (str): Name of FPGA class
            vcxo (int,float,rangec,arb_sourcec): Fixed VCXO value, range, or arb_source
            solver (str): Solver name (gekko, cplex)

        Raises:
            Exception: Unknown solver
            Exception: GEKKO Solver not installed
            Exception: arb_source requires CPLEX solver
        """
        if solver:
            self.solver = solver
        if self.solver == "gekko":
            if not solvers.gekko_solver:
                raise Exception("GEKKO Solver not installed")
            model = solvers.GEKKO(remote=False)
        elif self.solver == "CPLEX":
            if not solvers.cplex_solver:
                raise Exception("CPLEX Solver not installed")
            model = solvers.CpoModel()
        else:
            raise Exception(f"Unknown solver {self.solver}")

        self.model = model
        self.vcxo = vcxo
        self._plls = []
        self._plls_sysref = []

        # Validate arb_source compatibility with solver
        if isinstance(vcxo, arb_sourcec) and self.solver == "gekko":
            raise Exception(
                "arb_source type requires CPLEX solver. "
                "Either use solver='CPLEX' or use adijif.types.range "
                "for discrete values."
            )

        # FIXME: Do checks

        self.converter: Union[convc, List[convc]] = []
        if isinstance(conv, list):
            for c in conv:
                self.converter.append(
                    eval(f"adijif.{c}(self.model,solver=self.solver)")  # noqa: S307
                )
        else:
            self.converter: convc = eval(  # noqa: S307
                f"adijif.{conv}(self.model,solver=self.solver)"
            )
        self.clock = eval(f"adijif.{clk}(self.model,solver=self.solver)")  # noqa: S307
        self.fpga = eval(f"adijif.{fpga}(self.model,solver=self.solver)")  # noqa: S307
        self.vcxo = vcxo

        # TODO: Add these constraints to solver options
        self.configs_to_find = 1
        self.sysref_sample_clock_ratio = 16
        self.sysref_min_div = 4
        self.sysref_max_div = 2**14

        # self.solver_options = [
        #     "minlp_maximum_iterations 1000",  # minlp iterations with integer solution
        #     "minlp_max_iter_with_int_sol 100",  # treat minlp as nlp
        #     "minlp_as_nlp 0",  # nlp sub-problem max iterations
        #     "nlp_maximum_iterations 5000",  # 1 = depth first, 2 = breadth first
        #     "minlp_branch_method 1",  # maximum deviation from whole number
        #     "minlp_integer_tol 0.05",  # covergence tolerance
        #     "minlp_gap_tol 0.01",
        # ]
        self.solver_options = [
            "minlp_maximum_iterations 10000",  # minlp iterations with integer solution
            "minlp_max_iter_with_int_sol 100",  # treat minlp as nlp
            "minlp_as_nlp 0",  # nlp sub-problem max iterations
            "nlp_maximum_iterations 500",  # 1 = depth first, 2 = breadth first
            "minlp_branch_method 1",  # maximum deviation from whole number
            "minlp_integer_tol 0",  # covergence tolerance 0.000001
            "minlp_gap_tol 0.1",
        ]
        # self.solver_options = [
        #     "minlp_maximum_iterations 1000",  # minlp iterations with integer solution
        # ]

    def __del__(self) -> None:
        """Deconstructor: Cleanup system by clearing all leaf objects."""
        self.fpga = []
        if isinstance(self.converter, list):
            for c, _ in enumerate(self.converter):
                del self.converter[c]
        self.converter = []
        self.clock = []

    def _get_configs(self) -> Dict:
        """Collect extracted configurations from all components in system from solver.

        Returns:
            Dict: Dictionary containing all clocking configurations of all components
        """
        cfg = {"clock": self.clock.get_config(self.solution), "converter": []}

        c = self.converter if isinstance(self.converter, list) else [self.converter]
        for conv in c:
            if conv._nested:
                names = conv._nested
                for name in names:
                    clk_ref = cfg["clock"]["output_clocks"][
                        f"{self.fpga.name}_{name}_ref_clk"
                    ]["rate"]
                    cfg["fpga_" + name] = self.fpga.get_config(
                        solution=self.solution,
                        converter=getattr(conv, name),
                        fpga_ref=clk_ref,
                    )
                    cfg["converter"] = conv.get_config(self.solution)  # type: ignore
                    cfg["jesd_" + name] = getattr(conv, name).get_jesd_config(
                        self.solution
                    )
                    if getattr(conv, name).datapath:
                        cfg["datapath_" + name] = getattr(
                            conv, name
                        ).datapath.get_config()
            else:
                clk_ref = cfg["clock"]["output_clocks"][
                    f"{self.fpga.name}_{conv.name}_ref_clk"
                ]["rate"]
                cfg["fpga_" + conv.name] = self.fpga.get_config(
                    solution=self.solution, converter=conv, fpga_ref=clk_ref
                )
                cfg["converter_" + conv.name] = conv.get_config(self.solution)
                cfg["jesd_" + conv.name] = conv.get_jesd_config(self.solution)

        # Collect PLLs driving converter sampling clock configs
        for pll in self._plls:
            cfg["clock_ext_pll_" + pll.name] = pll.get_config(self.solution)

        # Collect PLLs driving sysref clocks configurations.
        for pll in self._plls_sysref:
            cfg["clock_ext_pll_sysref_" + pll.name] = pll.get_config(self.solution)
        return cfg

    def _filter_sysref(
        self,
        cnv_clocks: List,
        clock_names: List[str],
        convs: List[convc],
    ) -> Tuple[List, List[str]]:
        """Filter sysref clocks to remove duplicate constraints.

        Args:
            cnv_clocks (List): List of clock constraints
            clock_names (List[str]): List of clock names
            convs (List[convc]): List of converter objects

        Returns:
            List,List[str]: Touple with list of filter clocks and names

        Raises:
            Exception: Invalid shared sysref configuration
        """
        cnv_clocks_filters = []
        clock_names_filters = []
        if len(cnv_clocks) > 2 and self.use_common_sysref:
            ref = convs[0].multiframe_clock
            for conv in convs:
                if ref != conv.multiframe_clock:
                    raise Exception(
                        "SYSREF cannot be shared. "
                        + "Converters at different LMFCs."
                        + "\nSet use_common_sysref to False "
                        + "for current rates"
                    )

            for i, clk in enumerate(cnv_clocks):
                # 1,3,5,... are sysrefs. Keep first 1
                if i / 2 == int(i / 2) or i == 1:
                    cnv_clocks_filters.append(clk)
                    clock_names_filters.append(clock_names[i])
        else:
            cnv_clocks_filters = cnv_clocks
            clock_names_filters = clock_names
        return cnv_clocks_filters, clock_names_filters

    def _solve_gekko(self) -> None:
        """Call gekko solver API."""
        # Set up solver
        self.model.solver_options = self.solver_options
        self.model.options.SOLVER = 1  # APOPT solver
        folder = self.model._path
        # self.model.options.SOLVER = 3  # 1 APOPT, 2 BPOPT, 3 IPOPT
        # self.model.options.IMODE = 5   # simultaneous estimation
        try:
            self.model.solve(disp=self.Debug_Solver, debug=True)
            self.model.cleanup()
        finally:
            if os.path.isdir(folder) and os.path.isfile(
                os.path.join(folder, "apopt.opt")
            ):
                print("DELETE: " + folder)
                # shutil.rmtree(folder)

    def _solve_cplex(self) -> None:
        """Call CPLEX solver API."""
        # Set up solver
        ll = "Normal" if self.Debug_Solver else "Quiet"
        wl = 0  # WarningLevel 0-off 3-all warnings
        # self.model.export_model()
        self.solution = self.model.solve(LogVerbosity=ll, WarningLevel=wl)
        # self.solution.print_solution()
        if not self.solution.is_solution():
            raise Exception("No solution found")

    def solve(self) -> Dict:
        """Defined clocking requirements in Solver model and start solvers routine.

        Returns:
            Dict: Dictionary containing all clocking configuration for all components

        Raises:
            Exception: FPGA and Converter disabled
            Exception: Solver invalid
        """
        if not self.enable_converter_clocks and not self.enable_fpga_clocks:
            raise Exception("Converter and/or FPGA clocks must be enabled")

        # Reset objectives
        self.fpga._objectives = []
        self.clock._objectives = []
        self._objectives = []

        clock_names: List[str] = []
        config = {}
        if self.enable_converter_clocks:
            convs: List[convc] = (
                self.converter if isinstance(self.converter, list) else [self.converter]
            )

            # Setup clock chip
            self.clock._setup(self.vcxo)
            self.fpga.configs = []  # reset
            serdes_used_tx: int = 0
            serdes_used_rx: int = 0
            sys_refs = []  # DEBUG ONLY
            sys_ref_names = []  # DEBUG ONLY

            # Setup external sysref PLLs
            for pll in self._plls_sysref:
                if not isinstance(pll._ref, clockc):
                    pll._setup(pll._ref)

            for conv in convs:
                if conv._nested:  # MxFE, Transceivers
                    for name in conv._nested:
                        ctype = getattr(conv, name).converter_type.lower()
                        if ctype == "adc":
                            serdes_used_rx += getattr(conv, name).L
                        elif ctype == "dac":
                            serdes_used_tx += getattr(conv, name).L
                        else:
                            raise Exception(f"Unknown converter type {ctype}")
                else:
                    ctype = conv.converter_type.lower()
                    if ctype == "adc":
                        serdes_used_rx += conv.L
                    elif ctype == "dac":
                        serdes_used_tx += conv.L
                    else:
                        raise Exception(f"Unknown converter type: {ctype}")

                if (
                    serdes_used_rx > self.fpga.max_serdes_lanes
                    or serdes_used_tx > self.fpga.max_serdes_lanes
                ):
                    raise Exception(
                        "Max SERDES lanes exceeded. {} only available".format(
                            self.fpga.max_serdes_lanes
                        )
                    )

                # Check to make sure static configurations are in range
                conv.validate_config()

                # Check if we are using the same converter name
                if conv.name + "_ref_clk" in config:
                    raise Exception("Duplicate converter names found")

                # Setup sysref generator is separate (ADF4030) and driven by clock-chip
                for pll_sr in self._plls_sysref:
                    if isinstance(pll._ref, clockc):
                        for name in pll_sr._connected_to_output:
                            if name == conv.name:
                                # Get clock from clockchip
                                if conv._nested:
                                    raise Exception("Nested converters not supported")
                            else:
                                config, clock_names = self._get_ref_clock(
                                    pll_sr, config, clock_names
                                )
                                pll_sr._setup(config[pll_sr.name + "_ref_clk"])

                # Ask clock chip for converter ref
                config, clock_names = self._get_ref_clock(conv, config, clock_names)
                config, sys_ref_names = self._get_sysref_clock(
                    conv, config, sys_ref_names
                )

                # Setup converter
                clks = conv.get_required_clocks()  # type: ignore
                if not conv._nested:
                    assert len(clks) == 2, "Converter must have 2 clocks"

                # Check if converter uses external PLL
                for pll in self._plls:
                    if pll._connected_to_output == conv.name:
                        # Give clock chip output as PLL's reference
                        pll._setup(config[conv.name + "_ref_clk"])
                        # Give PLL's output as converter's reference
                        config[conv.name + "_ref_clk_from_ext_pll"] = (
                            pll._get_clock_constraint(
                                conv.name + "_ref_clk_from_ext_pll"
                            )
                        )
                        self.clock._add_equation(
                            config[conv.name + "_ref_clk_from_ext_pll"] == clks[0]
                        )

                # Connect converter to clock chip direct if no external PLL is used
                if all([conv.name != pll._connected_to_output for pll in self._plls]):
                    self.clock._add_equation(config[conv.name + "_ref_clk"] == clks[0])

                # Converter sysref
                sys_refs = self._apply_sysref_constraint(conv, clks, config, sys_refs)

                # Determine if separate device clock / link clock output is needed
                need_separate_link_clock = True  # Let solver decide

                # Ask clock chip for fpga ref
                config, clock_names = self._get_ref_clock_fpga(
                    conv, config, clock_names, need_separate_link_clock
                )

                # Setup fpga
                if conv._nested:
                    names = conv._nested
                    for name in names:
                        if need_separate_link_clock:
                            self.fpga.get_required_clocks(
                                getattr(conv, name),
                                config[name + "_fpga_ref_clk"],
                                config[name + "_fpga_device_clk"],
                            )
                        else:
                            self.fpga.get_required_clocks(
                                getattr(conv, name),
                                config[name + "_fpga_ref_clk"],
                            )
                else:
                    if need_separate_link_clock:
                        self.fpga.get_required_clocks(
                            conv,
                            config[conv.name + "_fpga_ref_clk"],
                            config[conv.name + "_fpga_device_clk"],
                        )
                    else:
                        self.fpga.get_required_clocks(
                            conv, config[conv.name + "_fpga_ref_clk"]
                        )

            # self.clock._clk_names = clock_names
            # FIXME: THIS IS TEMP HACK TO TEST
            # if self.plls_sysref:
            #     self._plls_sysref[0]._clk_names = sys_ref_names

            # Add post constraints
            if self.solver == "CPLEX":
                objectives = []
                self.clock._add_objective(sys_refs)
                for o in [self.fpga._objectives, self.clock._objectives]:
                    if o:
                        if isinstance(o, list):
                            objectives += o
                        else:
                            objectives.append(o)

                if objectives:
                    if len(self.fpga._objectives) > 1:
                        self.model.add(self.model.minimize_static_lex(objectives))
                    else:
                        self.model.minimize(objectives[0])

        # Set up solver
        if self.solver == "gekko":
            self._solve_gekko()
        elif self.solver == "CPLEX":
            self._solve_cplex()
        else:
            raise Exception("Unknown solver {}".format(self.solver))

        # Organize data
        return self._get_configs()

    def determine_clocks(self) -> List:
        """Defined clocking requirements and search over all possible dividers.

        Raises:
            Exception: No valid configurations found

        Returns:
            List: List of valid configurations for all clocking components
        """
        # Extract dependent rates from converter
        rates = self.converter.device_clock_available()  # type: ignore

        out = []
        for rate in rates:
            rate = np.array(rate, dtype=int)

            # Search across clock chip settings for supported modes
            clk_configs = self.clock.find_dividers(
                self.vcxo, rate, find=self.configs_to_find
            )

            # Find FPGA PLL settings that meet Lane rate
            # requirements based on available reference clocks
            valid_clock_configs = []
            for clk_config in clk_configs:
                refs = self.clock.list_available_references(clk_config)
                for ref in refs:
                    try:
                        info = self.fpga.determine_pll(
                            self.converter.bit_clock, ref  # type: ignore
                        )
                        break
                    except:  # noqa: B036, B001
                        ref = False
                        continue

                if ref:
                    clk_config["fpga_pll_config"] = info
                    valid_clock_configs.append(clk_config)

            if not valid_clock_configs:
                continue
                # raise Exception("No valid configurations possible for FPGA")

            # Check available output dividers for sysref required
            complete_clock_configs = []
            for clk_config in valid_clock_configs:
                refs = self.clock.list_available_references(clk_config)
                try:
                    sysref_rate = self._determine_sysref(refs)
                    clk_config["sysref_rate"] = sysref_rate
                    complete_clock_configs.append(clk_config)
                except:  # noqa: B036, B001, S112
                    continue

            if not complete_clock_configs:
                continue
                # raise Exception("No valid configurations possible for sysref")
            out.append({"Converter": rate, "ClockChip": complete_clock_configs})

        if not out:
            raise Exception("No valid configurations possible converter sample rate")

        return out

    def _determine_sysref(
        self, refs: Union[List[int], List[float]]
    ) -> Union[int, float]:
        """Find possible sysrefs based on required clocks and JESD configs.

        Args:
            refs (List[int],List[float]): List of system device clocks

        Raises:
            Exception: No valid configurations found

        Returns:
            int/float: Sysref rate in samples per second
        """
        lmfc = self.converter.multiframe_clock  # type: ignore
        div = self.sysref_min_div

        while div < self.sysref_max_div:
            sysref_rate = lmfc / div
            if sysref_rate in refs:
                # vco_div = cfs[0]['vco']/sysref/cfs[0]['m1']
                # print("FOUND")
                break
            else:
                sysref_rate = False
            div *= 2

        if not sysref_rate:
            raise Exception("No possible sysref found")

        return sysref_rate



"""Common class for all JIF components."""

from typing import List, Union

from adijif.solvers import CpoModel  # noqa: BLK100
from adijif.solvers import GK_Operators  # noqa: BLK100
from adijif.solvers import GEKKO, CpoExpr, GK_Intermediate, GKVariable


class core:
    """Common class for all JIF components.

    This is the central point for all classes in module

    """

    solver = "CPLEX"  # "CPLEX"

    _objectives = []

    def _add_objective(
        self,
        objective: List[Union[GKVariable, GK_Intermediate, GK_Operators, CpoExpr]],
    ) -> None:
        if isinstance(objective, list):
            self._objectives += objective
        else:
            self._objectives.append(objective)

    def __init__(
        self, model: Union[GEKKO, CpoModel] = None, solver: str = None
    ) -> None:
        """Initalize clocking model.

        When usings the clocking models standalone, typically for
        validation flows, a solver model is created internally.
        For system level work, a shared model is passed.

        Args:
            model (GEKKO,CpoModel): Solver model
            solver (str): Solver name (gekko or CPLEX)

        Raises:
            Exception: If solver is not valid
        """
        self._saved_solution = None
        self._objectives = []
        self._solution = None
        self.configs = []  # type: List[dict]
        if hasattr(self, "_init_diagram"):
            self._init_diagram()
        if solver:
            self.solver = solver
        if self.solver == "gekko":
            if model:
                assert isinstance(
                    model, GEKKO
                ), "Input model must be of type gekko.GEKKO"
            else:
                model = GEKKO(remote=False)
        elif self.solver == "CPLEX":
            if model:
                assert isinstance(
                    model, CpoModel
                ), "Input model must be of type docplex.cp.model.CpoModel"
            else:
                model = CpoModel()
        else:
            raise Exception(f"Unknown solver {self.solver}")
        self.model = model



"""Xilinx FPGA clocking model."""

from typing import Dict, List, Optional, Union

from docplex.cp.modeler import if_then

from ...converters.converter import converter as conv
from ...solvers import (
    CpoIntVar,
    CpoSolveResult,
    GK_Intermediate,
    GK_Operators,
    GKVariable,
)
from .bf import xilinx_bf
from .sevenseries import SevenSeries as SSTransceiver
from .ultrascaleplus import UltraScalePlus as USPTransceiver
from .xilinx_draw import xilinx_draw


class xilinx(xilinx_bf, xilinx_draw):
    """Xilinx FPGA clocking model.

    This model captures different limitations of the Xilinx
    PLLs and interfaces used for JESD.

    Currently only Zynq 7000 devices have been fully tested.
    """

    name = "Xilinx-FPGA"

    favor_cpll_over_qpll = False
    minimize_fpga_ref_clock = False

    """Force generation of separate device clock from the clock chip. In many
    cases, the ref clock and device clock can be the same."""
    force_separate_device_clock: bool = False

    """Constrain reference clock to be specific values. Options:
    - CORE_CLOCK: Make reference clock the same as the core clock (LR/40 or LR/66)
    - CORE_CLOCK_DIV2: Make reference clock the same as the core clock divided by 2
    - Unconstrained: No constraints on reference clock. Simply meet PLL constraints
    """
    _ref_clock_constraint = "CORE_CLOCK"

    max_serdes_lanes = 24

    hdl_core_version = 2.1

    available_speed_grades = [-1, -2, -3]
    speed_grade = -2

    transceiver_voltage = 800

    ref_clock_min = -1  # Not set
    ref_clock_max = -1  # Not set

    available_fpga_packages = [
        "Unknown",
        "RF",
        "FL",
        "FF",
        "FB",
        "HC",
        "FH",
        "CS",
        "CP",
        "FT",
        "FG",
        "SB",
        "RB",
        "RS",
        "CL",
        "SF",
        "BA",
        "FA",
    ]
    fpga_package = "FB"

    available_fpga_families = ["Unknown", "Artix", "Kintex", "Virtex", "Zynq"]
    fpga_family = "Zynq"

    available_transceiver_types = ["GTXE2"]
    transceiver_type = "GTXE2"

    def trx_gen(self) -> int:
        """Get transceiver generation (2,3,4).

        Returns:
            int: generation of transceiver
        """
        return int(self.transceiver_type[-1])

    def trx_variant(self) -> str:
        """Get transceiver variant (GTX, GTH, GTY, ...).

        Returns:
            str: Transceiver variant
        """
        # return self.transceiver_type[:2]
        trxt = self.transceiver_type[:2]
        assert len(trxt) == 3
        return trxt

    def fpga_generation(self) -> str:
        """Get FPGA generation 7000, US, US+... based on transceiver type.

        Returns:
            str: FPGA generation

        Raises:
            Exception: Unknown transceiver generation
        """
        if self.trx_gen() == 2:
            return "7000"
        elif self.trx_gen() == 3:
            return "Ultrascale"
        elif self.trx_gen() == 4:
            return "Ultrascale+"
        elif self.trx_gen() == 5:
            return "Versal"
        raise Exception(f"Unknown transceiver generation {self.trx_gen()}")

    sys_clk_selections = [
        "XCVR_CPLL",
        "XCVR_QPLL0",
        "XCVR_QPLL1",
    ]
    sys_clk_select = "XCVR_QPLL1"

    _out_clk_selections = [
        # "XCVR_OUTCLK_PCS",
        # "XCVR_OUTCLK_PMA",
        "XCVR_REFCLK",
        "XCVR_REFCLK_DIV2",
        "XCVR_PROGDIV_CLK",
    ]
    _out_clk_select = [
        # "XCVR_OUTCLK_PCS",
        # "XCVR_OUTCLK_PMA",
        "XCVR_REFCLK",
        "XCVR_REFCLK_DIV2",
        "XCVR_PROGDIV_CLK",
    ]

    """ Force use of QPLL for transceiver source """
    force_qpll = False

    """ Force use of QPLL1 for transceiver source (GTHE3,GTHE4,GTYE4)"""
    force_qpll1 = False

    """ Force use of CPLL for transceiver source """
    force_cpll = False

    """ Force all transceiver sources to be from a single PLL quad.
        This will try to leverage the output dividers of the PLLs
    """
    force_single_quad_tile = False

    """ Request that clock chip generated device clock
        device clock == LMFC/40
        NOTE: THIS IS NOT FPGA REF CLOCK
    """
    request_device_clock = False

    _clock_names: List[str] = []

    """When PROGDIV, this will be set to the value of the divider"""
    _used_progdiv = {}

    """FPGA target Fmax rate use to determine link layer output rate"""
    target_Fmax = 500e6

    """Device and reference clock relation"""
    _device_clock_and_ref_clock_relation_options = [
        "NA",
        "ref_clock_2x_device_clock",
        "ref_clock_eq_device_clock",
    ]
    _device_clock_and_ref_clock_relation = "NA"

    @property
    def device_clock_and_ref_clock_relation(self) -> str:
        """Get device clock and reference clock relation.

        Device clock and reference clock relation can be set to:
        - NA: No relation
        - ref_clock_2x_device_clock: Reference clock is 2x device clock
        - ref_clock_eq_device_clock: Reference clock is equal to device clock

        Returns:
            str: Device clock and reference clock relation.
        """
        return self._device_clock_and_ref_clock_relation

    @device_clock_and_ref_clock_relation.setter
    def device_clock_and_ref_clock_relation(self, value: str) -> None:
        """Set device clock and reference clock relation.

        Device clock and reference clock relation can be set to:
        - NA: No relation
        - ref_clock_2x_device_clock: Reference clock is 2x device clock
        - ref_clock_eq_device_clock: Reference clock is equal to device clock

        Args:
            value (str): Device clock and reference clock relation.

        Raises:
            Exception: Invalid device clock and reference clock relation selection.
        """
        if value not in self._device_clock_and_ref_clock_relation_options:
            raise Exception(
                f"Invalid device_clock_and_ref_clock_relation {value}, "
                + f"options are {self._device_clock_and_ref_clock_relation_options}"
            )
        self._device_clock_and_ref_clock_relation = value

    # """Require generation of separate clock specifically for link layer"""
    # requires_separate_link_layer_out_clock = True

    """Require generation of separate core clock (LR/40 or LR/66)"""
    requires_core_clock_from_device_clock = False

    _device_clock_source_options = ["external", "link_clock", "ref_clock"]
    _device_clock_source = ["external"]

    configs = []  # type: ignore
    _transceiver_models = {}  # type: ignore

    @property
    def device_clock_source(self) -> str:
        """Get device clock source.

        Device clock source can be set to:
        - external: External clock from clock chip
        - link_clock: Link layer clock is reused for as device clock
        - ref_clock: Reference clock is reused for device clock

        Returns:
            str: Device clock source.
        """
        return self._device_clock_source

    @device_clock_source.setter
    def device_clock_source(self, value: Union[str, List[str]]) -> None:
        """Set device clock source.

        Device clock source can be set to:
        - external: External clock
        - link_clock: Link layer clock
        - ref_clock: Reference clock

        Args:
            value (str, List[str]): Device clock source.

        Raises:
            Exception: Invalid device clock source selection.
        """
        if not isinstance(value, list):
            value = [value]
        for item in value:
            if item not in self._device_clock_source_options:
                raise Exception(
                    f"Invalid device clock source {item}, "
                    + f"options are {self._device_clock_source_options}"
                )
        self._device_clock_source = value

    @property
    def ref_clock_constraint(self) -> str:
        """Get reference clock constraint.

        Reference clock constraint can be set to:
        - CORE_CLOCK: Make reference clock the same as the core clock (LR/40 or LR/66)
        - CORE_CLOCK_DIV2: Make reference clock the same as the core clock divided by 2
        - Unconstrained: No constraints on reference clock. Simply meet PLL constraints

        Returns:
            str: Reference clock constraint.
        """
        return self._ref_clock_constraint

    _ref_clock_constraint_options = ["CORE_CLOCK", "CORE_CLOCK_DIV2", "Unconstrained"]

    @ref_clock_constraint.setter
    def ref_clock_constraint(self, value: str) -> None:
        """Set reference clock constraint.

        Reference clock constraint can be set to:
        - CORE_CLOCK: Make reference clock the same as the core clock (LR/40 or LR/66)
        - CORE_CLOCK_DIV2: Make reference clock the same as the core clock divided by 2
        - Unconstrained: No constraints on reference clock. Simply meet PLL constraints

        Args:
            value (str): Reference clock constraint.

        Raises:
            Exception: Invalid ref_clock_constraint selection.
        """
        if value not in self._ref_clock_constraint_options:
            raise Exception(
                f"Invalid ref_clock_constraint {value}, "
                + f"options are {self._ref_clock_constraint_options}"
            )
        self._ref_clock_constraint = value

    @property
    def out_clk_select(self) -> Union[int, float]:
        """Get current PLL clock output mux options for link layer clock.

        Valid options are: "XCVR_REFCLK", "XCVR_REFCLK_DIV2", "XCVR_PROGDIV_CLK"
        If a list of these is provided, the solver will determine one to use.

        Returns:
            str, list(str): Mux selection for link layer clock.
        """
        return self._out_clk_select

    @out_clk_select.setter
    def out_clk_select(self, value: Union[str, List[str]]) -> None:
        """Set current PLL clock output mux options for link layer clock.

        Valid options are: "XCVR_REFCLK", "XCVR_REFCLK_DIV2", "XCVR_PROGDIV_CLK"
        If a list of these is provided, the solver will determine one to use.

        Args:
            value (str, List[str]): Mux selection for link layer clock.

        Raises:
            Exception: Invalid out_clk_select selection.
        """
        if isinstance(value, list):
            for item in value:
                if item not in self._out_clk_selections:
                    raise Exception(
                        f"Invalid out_clk_select {item}, "
                        + f"options are {self._out_clk_selections}"
                    )
        elif isinstance(value, dict):
            for converter in value:
                if not isinstance(converter, conv):
                    raise Exception("Keys of out_clk_select but be of type converter")
                if value[converter] not in self._out_clk_selections:
                    raise Exception(
                        f"Invalid out_clk_select {value[converter]}, "
                        + f"options are {self._out_clk_selections}"
                    )
        elif value not in self._out_clk_selections:  # str
            raise Exception(
                f"Invalid out_clk_select {value}, "
                + f"options are {self._out_clk_selections}"
            )

        self._out_clk_select = value

    @property
    def _ref_clock_max(self) -> int:
        """Get maximum reference clock for config.

        Returns:
            int: Rate in samples per second.

        Raises:
            Exception: Unsupported transceiver type configured.
        """
        # https://www.xilinx.com/support/documentation/data_sheets/ds191-XC7Z030-XC7Z045-data-sheet.pdf # noqa: B950
        if self.transceiver_type == "GTXE2":
            if str(self.speed_grade) == "-3E":
                return 700000000
            else:
                return 670000000
        else:
            raise Exception(
                f"Unknown ref_clock_max for transceiver type {self.transceiver_type}"
            )
            # raise Exception(f"Unknown transceiver type {self.transceiver_type}")

    @property
    def _ref_clock_min(self) -> int:
        """Get minimum reference clock for config.

        Returns:
            int: Rate in samples per second.

        Raises:
            Exception: Unsupported transceiver type configured.
        """
        # https://www.xilinx.com/support/documentation/data_sheets/ds191-XC7Z030-XC7Z045-data-sheet.pdf # noqa: B950
        if self.transceiver_type == "GTXE2":
            return 60000000
        else:
            raise Exception(
                f"Unknown ref_clock_min for transceiver type {self.transceiver_type}"
            )
            # raise Exception(f"Unknown transceiver type {self.transceiver_type}")

    _available_dev_kit_names = [
        "zcu102",
        "zc706",
        "vcu118",
        "adsy1100",
    ]

    def setup_by_dev_kit_name(self, name: str) -> None:
        """Configure object based on board name. Ex: zc706, zcu102.

        Args:
            name (str): Name of dev kit. Ex: zc706, zcu102

        Raises:
            Exception: Unsupported board requested.

        """
        if name.lower() == "zc706":
            self.transceiver_type = "GTXE2"
            self.fpga_family = "Zynq"
            self.fpga_package = "FF"
            self.speed_grade = -2
            self.ref_clock_min = 60000000
            self.ref_clock_max = 670000000
            self.max_serdes_lanes = 8
            # default PROGDIV not available
            o = self._out_clk_selections.copy()
            del o[o.index("XCVR_PROGDIV_CLK")]
            self._out_clk_selections = o
            self._out_clk_select = o
        elif name.lower() == "zcu102":
            self.transceiver_type = "GTHE4"
            self.fpga_family = "Zynq"
            self.fpga_package = "FF"
            self.speed_grade = -2
            self.ref_clock_min = 60000000
            self.ref_clock_max = 820000000
            self.max_serdes_lanes = 8
        elif name.lower() == "vcu118":
            # XCVU9P-L2FLGA2104
            self.transceiver_type = "GTYE4"
            self.fpga_family = "Virtex"
            self.fpga_package = "FL"
            self.speed_grade = -2
            self.ref_clock_min = 60000000
            self.ref_clock_max = 820000000
            self.max_serdes_lanes = 24
        elif name.lower() == "adsy1100":
            # ADI VPX Module
            # VU11P: xcvu11p-flgb2104-2-i
            self.transceiver_type = "GTYE4"
            self.fpga_family = "Virtex"
            self.fpga_package = "FL"
            self.speed_grade = -2
            self.ref_clock_min = 60000000  # NEED TO VERIFY
            self.ref_clock_max = 820000000  # NEED TO VERIFY
            self.max_serdes_lanes = 24  # Connected to AD9084
        else:
            raise Exception(f"No boardname found in library for {name}")
        self.name = name

    def determine_pll(self, bit_clock: int, fpga_ref_clock: int) -> Dict:
        """Determine if configuration is possible with CPLL or QPLL.

        CPLL is checked first and will check QPLL if that case is
        invalid.

        This is only used for brute-force implementations.

        Args:
            bit_clock (int): Equivalent to lane rate in bits/second
            fpga_ref_clock (int): System reference clock

        Returns:
            Dict: Dictionary of PLL configuration
        """
        try:
            info = self.determine_cpll(bit_clock, fpga_ref_clock)
        except:  # noqa: B001
            info = self.determine_qpll(bit_clock, fpga_ref_clock)
        return info

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order

        Raises:
            Exception: Clock have not been enumerated aka get_required_clocks not
                not called yet.
        """
        if not self._clock_names:
            raise Exception(
                "get_required_clocks must be run to generated"
                + " dependent clocks before names are available"
            )
        return self._clock_names

    def get_config(
        self,
        converter: conv,
        fpga_ref: Union[float, int],
        solution: Optional[CpoSolveResult] = None,
    ) -> Union[List[Dict], Dict]:
        """Extract configurations from solver results.

        Collect internal FPGA configuration and output clock definitions.

        Args:
            converter (conv): Converter object connected to FPGA who config is
                collected
            fpga_ref (int or float): Reference clock generated for FPGA for specific
                converter
            solution (CpoSolveResult): CPlex solution and only needed for CPlex
                solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration.

        Raises:
            Exception: Invalid PLL configuration.
        """
        out = []
        if solution:
            self.solution = solution

        self._saved_solution = solution  # needed for draw

        for config in self.configs:
            pll_config: Dict[str, Union[str, int, float]] = {}

            # Filter out other converters
            # FIXME: REIMPLEMENT BETTER
            if converter.name + "_use_cpll" not in config.keys():
                print("Continued")
                continue

            pll_config = self._transceiver_models[converter.name].get_config(
                config, converter, fpga_ref
            )
            # cpll = self._get_val(config[converter.name + "_use_cpll"])
            # qpll = self._get_val(config[converter.name + "_use_qpll"])
            # if converter.name + "_use_qpll1" in config.keys():
            #     qpll1 = self._get_val(config[converter.name + "_use_qpll1"])
            # else:
            #     qpll1 = False

            # SERDES output mux
            if pll_config["type"] == "cpll":
                pll_config["sys_clk_select"] = "XCVR_CPLL"
            elif pll_config["type"] == "qpll":
                pll_config["sys_clk_select"] = "XCVR_QPLL0"
            elif pll_config["type"] == "qpll1":
                pll_config["sys_clk_select"] = "XCVR_QPLL1"
            else:
                raise Exception("Invalid PLL type")

            pll_config["out_clk_select"] = self._get_val(
                config[converter.name + "_out_clk_select"]
            )
            pll_config["out_clk_select"] = self._out_clk_selections[
                pll_config["out_clk_select"]
            ]

            if pll_config["out_clk_select"] == "XCVR_PROGDIV_CLK":
                progdiv = self._get_val(config[converter.name + "_progdiv_times2"])
                if float(int(progdiv / 2) * 2) != float(progdiv):
                    raise Exception(
                        f"PROGDIV divider {progdiv} is not an integer, "
                        + "this is not supported by the solver"
                    )
                progdiv = progdiv / 2
                progdiv = int(progdiv)

                # Check if we need progdiv really
                if progdiv == 1 and "XCVR_REFCLK" in self._out_clk_select:
                    # Change to XCVR_REFCLK
                    pll_config["out_clk_select"] = "XCVR_REFCLK"
                elif progdiv == 2 and "XCVR_REFCLK_DIV2" in self._out_clk_select:
                    # Change to XCVR_REFCLK_DIV2
                    pll_config["out_clk_select"] = "XCVR_REFCLK_DIV2"
                else:
                    if progdiv not in self._get_progdiv():
                        raise Exception(
                            f"PROGDIV divider {progdiv} not available for "
                            + f"transceiver type {self.transceiver_type}"
                        )
                    pll_config["progdiv"] = progdiv

            source = self._get_val(config[converter.name + "_device_clock_source"])
            pll_config["device_clock_source"] = self._device_clock_source_options[
                source
            ]

            if converter.jesd_class == "jesd204b":
                sps = converter.L * 32 / (converter.M * converter.Np)
            elif converter.jesd_class == "jesd204c":
                sps = converter.L * 64 / (converter.M * converter.Np)
            else:
                raise Exception("Invalid JESD class")
            pll_config["transport_samples_per_clock"] = sps

            out.append(pll_config)

        if len(out) == 1:
            out = out[0]  # type: ignore
        return out

    def _get_conv_prop(
        self, conv: conv, prop: Union[str, dict]
    ) -> Union[int, float, str]:
        """Helper to extract nested properties if present.

        Args:
            conv (conv): Converter object
            prop (str,dict): Property to extract

        Raises:
            Exception: Converter does not have property

        Returns:
            Union[int,float,str]: Value of property
        """
        if isinstance(prop, dict):
            if conv not in prop:
                raise Exception(f"Converter {conv.name} not found in config")
            return prop[conv]
        return prop

    def _get_progdiv(self) -> Union[List[int], List[float]]:
        """Get programmable SERDES dividers for FPGA.

        Raises:
            Exception: PRODIV is not available for transceiver type.

        Returns:
            List[int,float]: Programmable dividers for FPGA
        """
        if self.transceiver_type in ["GTYE3", "GTHE3"]:
            return [1, 4, 5, 8, 10, 16, 16.5, 20, 32, 33, 40, 64, 66, 80, 100]
        elif self.transceiver_type in ["GTYE4", "GTHE4"]:
            return [
                1,
                4,
                5,
                8,
                10,
                16,
                16.5,
                20,
                32,
                33,
                40,
                64,
                66,
                80,
                100,
                128,
                132,
            ]
        else:
            raise Exception(
                "PROGDIV is not available for FPGA transciever type "
                + str(self.transceiver_type)
            )

    def _set_link_layer_requirements(
        self,
        converter: conv,
        fpga_ref: Union[int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar],
        config: Dict,
        link_out_ref: Union[
            int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar
        ] = None,
    ) -> Dict:
        """Set link layer constraints for downstream FPGA logic.

        The link layer is driven from the XCVR core which can route the
        following signals to the link layer input:
        - External Ref
        - External Ref / 2
        - {CPLL,QPLL0,QPLL1} / PROGDIV

        The link layer input has a hard requirement that is must be at:
        - JESD204B: lane rate (bit clock) / 40 or lane rate (bit clock) / 80
        - JESD204C: lane rate (bit clock) / 66

        The link layer output rate will be equivalent to sample clock / N, where
        N is an integer. Note the smaller N, the more channeling it can be to
        synthesize the design. This output clock can be separate or be the same
        as the XCVR reference.

        Based on this info, we set up the problem where we define N (sample per
        clock increase), and set the lane rate based on the current converter
        JESD config.

        Args:
            converter (conv): Converter object connected to FPGA
            fpga_ref (int or GKVariable): Reference clock generated for FPGA
            config (Dict): Dictionary of clocking rates and dividers for link
                layer
            link_out_ref (int or GKVariable): Reference clock generated for FPGA
                link layer output

        Returns:
            Dict: Dictionary of clocking rates extended with dividers for link
                layer

        Raises:
            Exception: Link layer output clock select invalid
        """
        if converter.jesd_class == "jesd204b":
            link_layer_input_rate = converter.bit_clock / 40
        elif converter.jesd_class == "jesd204c":
            link_layer_input_rate = converter.bit_clock / 66

        if isinstance(self.out_clk_select, dict):
            if converter not in self.out_clk_select.keys():
                raise Exception(
                    "Link layer out_clk_select invalid for converter " + converter.name
                )
            if isinstance(self.out_clk_select[converter], dict):
                out_clk_select = self.out_clk_select[converter].copy()
            else:
                out_clk_select = self.out_clk_select[converter]
        else:
            out_clk_select = self.out_clk_select

        out_clk_select_options = []
        if "XCVR_REFCLK" in out_clk_select:
            out_clk_select_options.append(0)
        if "XCVR_REFCLK_DIV2" in out_clk_select:
            out_clk_select_options.append(1)
        if "XCVR_PROGDIV_CLK" in out_clk_select:
            out_clk_select_options.append(2)

        # Quick check for progdiv since it doesn't require the solver
        if "XCVR_PROGDIV_CLK" in out_clk_select and len(out_clk_select) == 1:
            progdiv = self._get_progdiv()
            div = converter.bit_clock / link_layer_input_rate
            if div not in progdiv:
                raise Exception(
                    f"Cannot use PROGDIV since required divider {div},"
                    + f" only available {progdiv}"
                )

        config[converter.name + "_out_clk_select"] = self._convert_input(
            out_clk_select_options, converter.name + "_out_clk_select"
        )

        if "XCVR_PROGDIV_CLK" in out_clk_select:
            progdiv = self._get_progdiv()
            progdiv_times2 = [i * 2 for i in progdiv]
            # Check that casting to int is valid
            for i in progdiv_times2:
                if i != int(i):
                    raise Exception(
                        f"PROGDIV times 2 divider {i} is not an integer, "
                        + "this is not supported by the solver"
                    )
            progdiv_times2 = [int(i) for i in progdiv_times2]
            config[converter.name + "_progdiv_times2"] = self._convert_input(
                progdiv_times2, converter.name + "_progdiv"
            )

        self._add_equation(
            [
                if_then(
                    config[converter.name + "_out_clk_select"] == 0,  # XCVR_REFCLK
                    fpga_ref == link_layer_input_rate,
                ),
                if_then(
                    config[converter.name + "_out_clk_select"] == 1,  # XCVR_REFCLK_DIV2
                    fpga_ref == link_layer_input_rate * 2,
                ),
            ]
        )

        if "XCVR_PROGDIV_CLK" in out_clk_select:
            self._add_equation(
                [
                    if_then(
                        config[converter.name + "_out_clk_select"]
                        == 2,  # XCVR_PROGDIV_CLK
                        converter.bit_clock * 2
                        == link_layer_input_rate
                        * config[converter.name + "_progdiv_times2"],
                    ),
                ]
            )

        return config, link_layer_input_rate

    def _setup_quad_tile(
        self,
        converter: conv,
        fpga_ref: Union[int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar],
        link_out_ref: Union[
            None, int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar
        ] = None,
    ) -> Dict:
        """Configure FPGA {Q/C}PLL tile.

        Args:
            converter (conv): Converter object(s) connected to FPGA
            fpga_ref (int,GKVariable, GK_Intermediate, GK_Operators, CpoIntVar):
                Reference clock for FPGA
            link_out_ref (None, int,GKVariable, GK_Intermediate, GK_Operators,
                CpoIntVar): Link layer output reference clock

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: Unsupported solver
        """
        # Add reference clock constraints
        self._add_equation(
            [fpga_ref >= self.ref_clock_min, fpga_ref <= self.ref_clock_max]
        )

        if converter.jesd_class == "jesd204b":
            core_clock = converter.bit_clock / 40
        else:
            core_clock = converter.bit_clock / 66

        if self.ref_clock_constraint == "CORE_CLOCK":
            self._add_equation([fpga_ref == core_clock])
        elif self.ref_clock_constraint == "CORE_CLOCK_DIV2":
            self._add_equation([fpga_ref == core_clock / 2])

        # Add transceiver
        config = {}
        if self.fpga_generation() == "7000":
            self._transceiver_models[converter.name] = SSTransceiver(
                parent=self,
                transceiver_type=self.transceiver_type,
                speed_grade=self.speed_grade,
            )
        elif self.fpga_generation() in ["Ultrascale", "Ultrascale+"]:
            self._transceiver_models[converter.name] = USPTransceiver(
                parent=self,
                transceiver_type=self.transceiver_type,
                speed_grade=self.speed_grade,
            )
        else:
            raise Exception(f"Unsupported FPGA generation {self.fpga_generation()}")

        # Handle force PLLs for nested devices and multiple converters
        force_cpll = False
        force_qpll = False
        force_qpll1 = False

        if isinstance(self.force_cpll, dict):
            if converter in self.force_cpll:
                force_cpll = self.force_cpll[converter]
        else:
            force_cpll = self.force_cpll

        if isinstance(self.force_qpll, dict):
            if converter in self.force_qpll:
                force_qpll = self.force_qpll[converter]
        else:
            force_qpll = self.force_qpll

        if isinstance(self.force_qpll1, dict):
            if converter in self.force_qpll1:
                force_qpll1 = self.force_qpll1[converter]
        else:
            force_qpll1 = self.force_qpll1

        self._transceiver_models[converter.name].force_cpll = force_cpll
        self._transceiver_models[converter.name].force_qpll = force_qpll
        if hasattr(self._transceiver_models[converter.name], "force_qpll1"):
            self._transceiver_models[converter.name].force_qpll1 = force_qpll1

        config = self._transceiver_models[converter.name].add_constraints(
            config, fpga_ref, converter
        )

        # Add constraints for link clock and transport clock
        # Link clock in must be lane rate / 40 or lane rate / 66
        # Link clock out and transport must be sample rate / SPS

        # Add constraints for link clock input
        config, link_layer_input_rate = self._set_link_layer_requirements(
            converter, fpga_ref, config, None
        )

        # Add device clock constraints
        # Device clock drives the output clock for the link layer and the
        # transport clock. It is sample_rates / SPS
        if converter.jesd_class == "jesd204b":
            sps = converter.L * 32 / (converter.M * converter.Np)
        elif converter.jesd_class == "jesd204c":
            sps = converter.L * 64 / (converter.M * converter.Np)

        device_clock_rate = converter.sample_clock / sps

        # Quick check for non-solver case(s)
        if (
            len(self._device_clock_source) == 1
            and self._device_clock_source[0] == "link_clock"
            and device_clock_rate != link_layer_input_rate
        ):
            raise Exception(
                "Configuration not possible if device_clock_source is link_clock\n"
                + f"Device clock rate {device_clock_rate} != Link layer input "
                + f"rate {link_layer_input_rate}"
            )
        # Options: "external", "link_clock", "ref_clock"
        dcs = []
        if "external" in self._device_clock_source:
            dcs.append(0)
        if "link_clock" in self._device_clock_source:
            dcs.append(1)
        if "ref_clock" in self._device_clock_source:
            dcs.append(2)

        config[converter.name + "_device_clock_source"] = self._convert_input(
            dcs, converter.name + "_device_clock_source"
        )
        self._add_equation(
            [
                if_then(
                    config[converter.name + "_device_clock_source"] == 0,  # external
                    link_out_ref == device_clock_rate,
                ),
                if_then(
                    config[converter.name + "_device_clock_source"]
                    == 1,  # link_clock_in
                    link_layer_input_rate == device_clock_rate,
                ),
                if_then(
                    config[converter.name + "_device_clock_source"] == 2,  # ref_clk
                    fpga_ref == link_out_ref,
                ),
            ]
        )

        # Add device clock and ref clock interaction
        if not self.device_clock_and_ref_clock_relation == "NA":
            if self.device_clock_and_ref_clock_relation == "ref_clock_2x_device_clock":
                self._add_equation([fpga_ref == 2 * link_out_ref])
            elif (
                self.device_clock_and_ref_clock_relation == "ref_clock_eq_device_clock"
            ):
                self._add_equation([fpga_ref == link_out_ref])
            else:
                raise Exception(
                    "Invalid device clock and reference clock \n"
                    + f"relation {self.device_clock_and_ref_clock_relation}"
                )

        return config

        # Add optimization to favor a single reference clock vs unique ref+device clocks
        config[converter.name + "single_clk"] = self._convert_input(
            [0, 1], converter.name + "single_clk"
        )
        if self.force_separate_device_clock:
            sdc = [1]
        else:
            sdc = [0, 1]
        config[converter.name + "two_clks"] = self._convert_input(
            sdc, converter.name + "two_clks"
        )
        self._add_equation(
            [
                config[converter.name + "single_clk"]
                + config[converter.name + "two_clks"]
                == 1,
            ]
        )
        # Favor single clock, this equation will be minimized
        v = (
            config[converter.name + "single_clk"]
            + 1000 * config[converter.name + "two_clks"]
        )
        self._add_objective(v)

        # Add constraints to meet sample clock
        if self.requires_core_clock_from_device_clock:
            if converter.jesd_class == "jesd204b":
                core_clock = converter.bit_clock / 40
            else:
                core_clock = converter.bit_clock / 66

            self._add_equation([core_clock == link_out_ref])

        else:
            possible_divs = []
            for samples_per_clock in [1, 2, 4, 8, 16]:
                if converter.sample_clock / samples_per_clock <= self.target_Fmax:
                    possible_divs.append(samples_per_clock)

            if len(possible_divs) == 0:
                raise Exception("Link layer output clock rate too high")

            config[converter.name + "_link_out_div"] = self._convert_input(
                possible_divs, converter.name + "_samples_per_clock"
            )

            self._add_equation(
                [
                    (
                        config[converter.name + "single_clk"] * fpga_ref
                        + config[converter.name + "two_clks"]
                        * link_out_ref
                        * config[converter.name + "_link_out_div"]
                    )
                    == converter.sample_clock
                ]
            )

        return config

    def get_required_clocks(
        self,
        converter: conv,
        fpga_ref: Union[int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar],
        link_out_ref: Union[
            int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar
        ] = None,
    ) -> List:
        """Get necessary clocks for QPLL/CPLL configuration.

        Args:
            converter (conv): Converter object of converter connected to
                FPGA
            fpga_ref (int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar):
                Abstract or concrete reference to FPGA reference clock
            link_out_ref (int or GKVariable): Reference clock generated for FPGA
                link layer output, also called device clock

        Returns:
            List: List of solver variables and constraints

        Raises:
            Exception: If solver is not valid
            Exception: Link layer out clock required
        """
        if self.ref_clock_min == -1 or self.ref_clock_max == -1:
            raise Exception("ref_clock_min or ref_clock_max not set")
        if "_get_converters" in dir(converter):
            converter = (
                converter._get_converters()  # type: ignore
            )  # Handle nested converters

        if not isinstance(converter, list):
            converter = [converter]  # type: ignore

        # if self.solver == "gekko":
        #     self.config = {
        #         "fpga_ref": self.model.Var(
        #             # integer=True,
        #             lb=self.ref_clock_min,
        #             ub=self.ref_clock_max,
        #             value=self.ref_clock_min,
        #         )
        #     }
        # elif self.solver == "CPLEX":
        #     # self.config = {
        #     #     "fpga_ref": integer_var(
        #     #         self.ref_clock_min, self.ref_clock_max, "fpga_ref"
        #     #     )
        #     # }
        #     pass
        # else:
        #     raise Exception(f"Unknown solver {self.solver}")

        # https://www.xilinx.com/support/documentation/user_guides/ug476_7Series_Transceivers.pdf # noqa: B950

        # clock_names = ["fpga_ref"]
        clock_names = []
        self.config = {}
        if self.force_single_quad_tile:
            raise Exception("force_single_quad_tile==1 not implemented")
        else:
            #######################
            # self.configs = []
            self.dev_clocks = []
            self.ref_clocks = []
            # obs = []
            for cnv in converter:  # type: ignore
                # rsl = self._get_conv_prop(
                #     cnv, self.requires_separate_link_layer_out_clock
                # )
                # if link_out_ref is None and rsl:
                #     raise Exception("Link layer out clock required")

                clock_names.append(cnv.name + "fpga_ref")
                # self.config[cnv.name+"fpga_ref"] = interval_var(
                #     self.ref_clock_min, self.ref_clock_max, name=cnv.name+"fpga_ref"
                # )
                self.config[cnv.name + "fpga_ref"] = fpga_ref
                self.ref_clocks.append(self.config[cnv.name + "fpga_ref"])
                if (
                    link_out_ref is not None
                ):  # self.requires_separate_link_layer_out_clock:
                    self.config[cnv.name + "link_out_ref"] = link_out_ref
                    self.ref_clocks.append(self.config[cnv.name + "link_out_ref"])
                    config = self._setup_quad_tile(
                        cnv,
                        self.config[cnv.name + "fpga_ref"],
                        self.config[cnv.name + "link_out_ref"],
                    )
                else:
                    config = self._setup_quad_tile(
                        cnv, self.config[cnv.name + "fpga_ref"]
                    )
                # Set optimizations
                # self.model.Obj(self.config[converter.name+"d"])
                # self.model.Obj(self.config[converter.name+"d_cpll"])
                # self.model.Obj(config[converter.name+"d_select"])
                if self.favor_cpll_over_qpll:
                    if self.solver == "gekko":
                        self.model.Obj(
                            -1 * config[cnv.name + "qpll_0_cpll_1"]
                        )  # Favor CPLL over QPLL
                    elif self.solver == "CPLEX":
                        self.model.maximize(config[cnv.name + "qpll_0_cpll_1"])
                        # obs.append(-1 * config[cnv.name + "qpll_0_cpll_1"])
                    else:
                        raise Exception(f"Unknown solver {self.solver}")

                self.configs.append(config)
                # FPGA also requires clock at device clock rate
                if self.request_device_clock:
                    self.dev_clocks.append(cnv.device_clock)
                    clock_names.append(cnv.name + "_fpga_device_clock")

        if self.minimize_fpga_ref_clock:
            if self.solver == "gekko":
                self.model.Obj(self.config[cnv.name + "fpga_ref"])
            elif self.solver == "CPLEX":
                # self.model.minimize_static_lex(obs + [self.config[converter.name+"fpga_ref"]]) # noqa: B950
                self.model.minimize(self.config[cnv.name + "fpga_ref"])  # noqa: B950
                # self.model.maximize(obs + self.config[converter.name+"fpga_ref"])
            else:
                raise Exception(f"Unknown solver {self.solver}")

        self._clock_names = clock_names

        # return [self.config["fpga_ref"]] + self.dev_clocks
        return self.ref_clocks + self.dev_clocks



"""Converter base meta class for all converter clocking models."""

from abc import ABCMeta, abstractmethod
from typing import List, Union

from adijif.converters.converter import converter


class adc(converter, metaclass=ABCMeta):
    """ADC base meta class used to enforce all ADC classes.

    This class should be inherited from for all ADCs and transceivers.

    """

    def _check_valid_internal_configuration(self) -> None:
        """Verify current internal clocking configuration for part is valid.

        Raises:
            Exception: Invalid clocking configuration
        """
        if self.decimation * self.sample_clock > self.converter_clock_max:
            raise Exception(
                "ADC rate too fast for configuration {}".format(
                    self.decimation * self.sample_clock
                )
            )
        if self.decimation * self.sample_clock < self.converter_clock_min:
            raise Exception(
                "ADC rate too slow for configuration {}".format(
                    self.decimation * self.sample_clock
                )
            )

    @property
    @abstractmethod
    def decimation_available(self) -> List[int]:
        """Decimation settings available.

        Must be a list or None

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    _decimation = 1

    @property
    def decimation(self) -> Union[int, float]:
        """Decimation between DAC and JESD framer. If none device is not an ADC.

        Generally a multiple of 2

        Returns:
            int: decimation value
        """
        return self._decimation

    @decimation.setter
    def decimation(self, value: int) -> None:
        """Decimation between DAC and JESD framer. If none device is not an ADC.

        Args:
            value (int): decimation

        Raises:
            Exception: decimation not an integer or not in range
        """
        if int(value) != value:
            raise Exception("decimation_available must be an integer")
        if value not in self.decimation_available:
            raise Exception("decimation_available not in range for device")
        self._decimation = value

    @property
    def converter_clock(self) -> Union[int, float]:
        """Get rate of converter in samples per second.

        Returns:
            float: converter clock rate in samples per second
        """
        return self.sample_clock * self.decimation



"""Streamlit testing for jesd mode selector page."""

# Locate the main app file
import pathlib
import sys
import time
from typing import Any, Optional

from streamlit.testing.v1 import AppTest

from adijif import converters

app_path = pathlib.Path(__file__).parent.parent.parent / "adijif" / "tools" / "explorer"
app_file_path = app_path / "main.py"

# Add app path to sys.path
sys.path.append(str(app_path))


# Webui helpers
def get_item(
    at: AppTest, t: str, label: str, debug: bool = False
) -> Optional[Any]:  # noqa: ANN401
    """Get a UI item by label or key.

    Args:
        at: AppTest instance
        t: Type of item to get
        label: Label or key to search for
        debug: Whether to print debug information

    Returns:
        Found item or None
    """
    for item in getattr(at, t):
        if debug:
            print("Found:", t, item.label, item.value, item.type, item.key)
            print(f"{item.label=}, {item.value=}, {item.type=}, {item.key=}")
            print("Looking for key:", label)
        if item.key == label or item.label == label:
            return item
    return None


def set_item(
    at: AppTest,
    t: str,
    label: str,
    value: Any,  # noqa: ANN401
    skip_warning_check: bool = False,
    print_warnings: bool = False,
    do_read_back: bool = False,
    debug: bool = False,
) -> Any:  # noqa: ANN401
    """Set a UI item value.

    Args:
        at: AppTest instance
        t: Type of item to set
        label: Label or key of item
        value: Value to set
        skip_warning_check: Whether to skip warning checks
        print_warnings: Whether to print warnings
        do_read_back: Whether to verify value after setting
        debug: Whether to print debug information

    Returns:
        The item that was set

    Raises:
        ValueError: If item is not found or warning occurs during setting
    """
    item = get_item(at, t, label, debug=debug)
    if item:
        # print(f"Set {label} to {value} for {t}")
        # print("Item details")
        # print(item)
        # print('----------')
        item.set_value(value).run()
        if do_read_back:
            itemr = get_item(at, t, label)
            print(f"Read back: {itemr.value}")
            options = itemr.options
            print(f"Options: {options}")
            assert itemr.value == value
        # print('#############')
        assert item.value == value
        if at.exception:
            print(item)
        assert not at.exception
        # if print_warnings:
        #     if at.session_state.warning:
        #         print(f"Warning: {at.session_state.warning}")
        # if not skip_warning_check:
        #     if at.session_state.warning:
        #         itemr = get_item(at, t, label)
        #         options = itemr.options
        #         print(f"Options: {options}")
        #         raise ValueError(f"Warning: {at.session_state.warning}")
    else:
        raise ValueError(f"Item {label} not found")
    return item


def test_jesdmodeselector_page_basic() -> None:
    """Test basic jesd mode selector page loading."""
    at = AppTest.from_file(app_file_path).run()

    # Navigate to the JESD Mode Selector page
    # sb = at.sidebar
    # set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(1)

    # Verify page title
    assert len(at.title) > 0
    assert at.title[0].value == "JESD204 Mode Selector"

    # Check that help button exists
    assert len(at.button) > 0
    help_button_found = False
    for button in at.button:
        if button.label == "Help":
            help_button_found = True
            break
    assert help_button_found, "Help button not found"


def test_jesdmodeselector_all_parts_load() -> None:
    """Test that all supported parts load without errors."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar

    # Navigate to the JESD Mode Selector page
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(1)

    parts = converters.supported_parts
    for part in parts:
        set_item(at, "selectbox", "Select a part", part)
        at.run()
        time.sleep(1)
        # Check if errors or warnings are present
        assert not at.exception, f"Exception occurred for part {part}"


def test_jesdmodeselector_adc_datapath_config() -> None:
    """Test ADC datapath configuration options."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    # Select an ADC part (ad9680 is a known ADC)
    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Verify expanders exist
    assert len(at.expander) > 0, "No expanders found"

    # Check for Datapath Configuration expander
    datapath_expander_found = False
    for expander in at.expander:
        if "Datapath Configuration" in str(expander.label):
            datapath_expander_found = True
            break
    assert datapath_expander_found, "Datapath Configuration expander not found"

    # Check for Units selector
    units_found = False
    for selectbox in at.selectbox:
        if selectbox.label == "Units":
            units_found = True
            assert selectbox.value == "GHz", "Default unit should be GHz"
            assert "Hz" in selectbox.options
            assert "kHz" in selectbox.options
            assert "MHz" in selectbox.options
            assert "GHz" in selectbox.options
            break
    assert units_found, "Units selector not found"

    assert not at.exception


def test_jesdmodeselector_converter_rate_units() -> None:
    """Test converter rate units selection and conversion."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Test different unit selections
    for unit in ["Hz", "kHz", "MHz", "GHz"]:
        set_item(at, "selectbox", "Units", unit)
        at.run()
        assert not at.exception, f"Exception occurred when selecting {unit}"

        # Verify converter rate input exists
        converter_rate_found = False
        for number_input in at.number_input:
            if f"Converter Rate ({unit})" in number_input.label:
                converter_rate_found = True
                break
        assert converter_rate_found, f"Converter Rate input not found for {unit}"


def test_jesdmodeselector_derived_settings() -> None:
    """Test that derived settings are displayed."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Check that dataframe with derived settings exists
    assert len(at.dataframe) > 0, "No dataframe found for derived settings"

    # Check that the first dataframe contains sample rate
    df_found = False
    for df in at.dataframe:
        if hasattr(df, "value") and df.value is not None:
            if "Sample Rate (MSPS)" in str(df.value):
                df_found = True
                break
    assert df_found, "Sample Rate derived setting not found"

    assert not at.exception


def test_jesdmodeselector_jesd_modes_display() -> None:
    """Test that JESD modes are displayed correctly."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Check that Configuration and JESD204 Modes subheaders exist
    assert len(at.subheader) >= 2, "Expected at least 2 subheaders"

    subheader_texts = [sh.value for sh in at.subheader]
    assert "Configuration" in subheader_texts, "Configuration subheader not found"
    assert "JESD204 Modes" in subheader_texts, "JESD204 Modes subheader not found"

    # Check for toggle to show only valid modes
    toggle_found = False
    for toggle in at.toggle:
        if "Show only valid modes" in toggle.label:
            toggle_found = True
            assert toggle.value is True, "Toggle should default to True"
            break
    assert toggle_found, "Show only valid modes toggle not found"

    assert not at.exception


def test_jesdmodeselector_valid_modes_toggle() -> None:
    """Test toggling between valid and all modes."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Find and toggle the "Show only valid modes" toggle
    toggle_item = None
    for toggle in at.toggle:
        if "Show only valid modes" in toggle.label:
            toggle_item = toggle
            break

    assert toggle_item is not None, "Show only valid modes toggle not found"

    # Toggle to show all modes
    toggle_item.set_value(False).run()
    assert not at.exception, "Exception occurred when toggling to show all modes"

    # Toggle back to show only valid modes
    toggle_item.set_value(True).run()
    assert not at.exception, "Exception occurred when toggling to show valid modes"


def test_jesdmodeselector_jesd_configuration_multiselect() -> None:
    """Test JESD configuration multiselect options."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Check that multiselect widgets exist for JESD parameters
    assert (
        len(at.multiselect) > 0
    ), "No multiselect widgets found for JESD configuration"

    # Common JESD parameters that might be present
    expected_params = ["M", "L", "N", "Np", "F", "S", "K", "HD", "CS"]

    # Check that at least some JESD parameters are available as multiselect
    found_params = []
    for ms in at.multiselect:
        if ms.label in expected_params:
            found_params.append(ms.label)

    assert len(found_params) > 0, "No JESD configuration parameters found"

    assert not at.exception


def test_jesdmodeselector_dac_part() -> None:
    """Test DAC part loading (if available)."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    # Try to find a DAC part (ad9136 is a known DAC)
    dac_parts = ["ad9136", "ad9144", "ad9172", "ad9081"]
    dac_found = False

    for part in dac_parts:
        if part in converters.supported_parts:
            try:
                set_item(at, "selectbox", "Select a part", part)
                at.run()
                dac_found = True

                # Note: Some parts might not have interpolation settings
                # so we just check that the page loads without errors
                assert not at.exception, f"Exception occurred for DAC part {part}"
                break
            except Exception:  # noqa: S110, S112
                continue

    # If no DAC parts found, that's okay - just note it
    if not dac_found:
        print("Note: No DAC parts with quick_configuration_modes found for testing")


def test_jesdmodeselector_diagram_adc() -> None:
    """Test that diagram is shown for ADC parts."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Check for Diagram expander
    diagram_expander_found = False
    for expander in at.expander:
        if "Diagram" in str(expander.label):
            diagram_expander_found = True
            break

    assert diagram_expander_found, "Diagram expander not found"

    # Check that an image is rendered (images are rendered within expanders)
    # We just verify the expander exists and no exception occurred
    assert not at.exception


def test_jesdmodeselector_data_editor_present() -> None:
    """Test that JESD modes section is present."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Check that JESD204 Modes subheader exists
    # (this indicates the modes section is rendered)
    jesd_modes_subheader_found = False
    for subheader in at.subheader:
        if "JESD204 Modes" in subheader.value:
            jesd_modes_subheader_found = True
            break

    assert jesd_modes_subheader_found, "JESD204 Modes subheader not found"

    # Check that toggle for valid modes exists
    toggle_found = False
    for toggle in at.toggle:
        if "Show only valid modes" in toggle.label:
            toggle_found = True
            break

    assert toggle_found, "Show only valid modes toggle not found"

    assert not at.exception


def test_jesdmodeselector_converter_rate_change() -> None:
    """Test changing converter rate and verify no errors."""
    at = AppTest.from_file(app_file_path).run()

    sb = at.sidebar
    set_item(sb, "radio", "Select a Tool", "JESD204 Mode Selector")
    time.sleep(0.5)

    set_item(at, "selectbox", "Select a part", "ad9680")
    at.run()

    # Find converter rate input
    for number_input in at.number_input:
        if "Converter Rate" in number_input.label:
            # Try setting different values
            original_value = number_input.value
            test_values = [0.5, 1.0, 2.0]

            for test_val in test_values:
                try:
                    number_input.set_value(test_val).run()
                    assert (
                        not at.exception
                    ), f"Exception when setting rate to {test_val}"
                except Exception:  # noqa: S110
                    # Some values might be out of range, which is okay
                    pass

            # Set back to original value
            try:
                number_input.set_value(original_value).run()
            except Exception:  # noqa: S110
                pass
            break



# flake8: noqa
# pytype: skip-file
import numpy as np

from adijif.clocks.clock import clock


class ad9528_bf(clock):
    """Brute force methods for calculating clocks

    These are currently meant for debug to compare against
    the solver solutions
    """

    def list_available_references(self, divider_set):
        """list_available_references: Based on config list possible
        references that can be generated based on VCO and output
        dividers
        """
        # Check input
        ref = {
            "m1": 3,
            "n2": 2,
            "vco": 3000000000,
            "r1": 24,
            "required_output_divs": np.array([1.0]),
        }
        for key in ref:
            if key not in divider_set:
                raise Exception(
                    "Input must be of type dict with fields: " + str(ref.keys())
                )
        return [
            divider_set["vco"] / divider_set["m1"] / div for div in self.d_available
        ]

    def find_dividers(self, vcxo, required_output_rates, find=3):
        if self.use_vcxo_double:
            vcxo *= 2

        mod = np.gcd.reduce(np.array(required_output_rates, dtype=int))
        configs = []

        for r1 in self.r1_available:
            pfd = vcxo / r1
            if pfd > self.pfd_max:
                continue
            for m1 in self.m1_available:
                for n2 in self.n2_available:
                    # RECHECK THIS. NOT WELL DOCUMENTED
                    vco = pfd * m1 * n2
                    # Check bounds and integer
                    if (
                        vco > self.vco_min
                        and vco < self.vco_max
                        and (vco / m1) % mod == 0
                    ):
                        required_output_divs = (vco / m1) / required_output_rates
                        if np.all(np.isin(required_output_divs, self.d_available)):
                            configs.append(
                                {
                                    "m1": m1,
                                    "vco": vco,
                                    "n2": n2,
                                    "r1": r1,
                                    "required_output_divs": required_output_divs,
                                }
                            )

        return configs



"""PLL parent metaclass to maintain consistency for all pll chips."""

from abc import ABCMeta
from typing import Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.common import core
from adijif.gekko_trans import gekko_translation


class pll(core, gekko_translation, metaclass=ABCMeta):
    """Parent metaclass for all pll chip classes."""

    # @property
    # @abstractmethod
    # def find_dividers(self) -> Dict:
    #     """Find all possible divider settings that validate config.

    #     Raises:
    #         NotImplementedError: Method not implemented
    #     """
    #     raise NotImplementedError  # pragma: no cover

    # @property
    # @abstractmethod
    # def list_available_references(self) -> List[int]:
    #     """Determine all references that can be generated.

    #     Based on config list possible references that can be generated
    #     based on VCO and output dividers

    #     Raises:
    #         NotImplementedError: Method not implemented
    #     """
    #     raise NotImplementedError  # pragma: no cover

    _connected_to_output = ""
    _connected_to_input = ""

    def _solve_gekko(self) -> bool:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            bool: Always False
        """
        self.model.options.SOLVER = 1  # APOPT solver
        self.model.solver_options = [
            "minlp_maximum_iterations 1000",  # minlp iterations with integer solution
            "minlp_max_iter_with_int_sol 100",  # treat minlp as nlp
            "minlp_as_nlp 0",  # nlp sub-problem max iterations
            "nlp_maximum_iterations 500",  # 1 = depth first, 2 = breadth first
            "minlp_branch_method 1",  # maximum deviation from whole number
            "minlp_integer_tol 0",  # covergence tolerance (MUST BE 0 TFC)
            "minlp_gap_tol 0.1",
        ]

        self.model.solve(disp=False)
        self.model.cleanup()
        return False

    # def _add_objective(self, sysrefs: List) -> None:
    #     pass

    def _solve_cplex(self) -> CpoSolveResult:
        if self._objectives:
            if len(self._objectives) == 1:
                self.model.maximize(self._objectives[0])
            else:
                self.model.add(self.model.maximize_static_lex(self._objectives))
        self.model.export_model()
        self.solution = self.model.solve(
            # Workers=1,
            # agent="local",
            # SearchType="DepthFirst",
            LogVerbosity="Verbose",
            # OptimalityTolerance=1e-12,
            # RelativeOptimalityTolerance=1e-12,
        )
        if self.solution.solve_status not in ["Feasible", "Optimal"]:
            raise Exception("Solution Not Found")
        return self.solution

    def solve(self) -> Union[None, CpoSolveResult]:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            [None,CpoSolveResult]: When cplex solver is used CpoSolveResult is returned

        Raises:
            Exception: If solver is not valid

        """
        if self.solver == "gekko":
            return self._solve_gekko()
        elif self.solver == "CPLEX":
            return self._solve_cplex()
        else:
            raise Exception(f"Unknown solver {self.solver}")



# flake8: noqa
import numpy as np

from adijif.clocks.clock import clock


class hmc7044_bf(clock):
    """Brute force methods for calculating clocks

    These are currently meant for debug to compare against
    the solver solutions
    """

    def list_available_references(self, divider_set):
        """list_available_references: Based on config list possible
        references that can be generated based on VCO and output
        dividers
        """
        # Check input
        ref = {
            "n2": 2,
            "vco": 3000000000,
            "r2": 24,
            "required_output_divs": np.array([1.0]),
        }
        for key in ref:
            if key not in divider_set:
                raise Exception(
                    "Input must be of type dict with fields: " + str(ref.keys())
                )
        return [divider_set["vco"] / div for div in self.d_available]

    def find_dividers(self, vcxo, rates, find=3):
        if self.use_vcxo_double:
            vcxo *= 2

        even = np.arange(2, 4096, 2, dtype=int)
        odivs = np.append([1, 3, 5], even)

        rates = np.array(rates)
        mod = np.gcd.reduce(np.array(rates, dtype=int))
        vcos = []
        configs = []

        # for n in range(self.n2_divider_min, self.n2_divider_max):
        for n in self.n2:
            # for r in range(self.r2_divider_min, self.r2_divider_max):
            for r in self.r2:
                # Check VCO in range and output clock a multiple of required reference
                f = vcxo * n / r
                if f >= self.vco_min and f <= self.vco_max:
                    # Check if required dividers for output clocks are in set
                    if f % mod == 0:
                        d = f / rates
                        if np.all(np.isin(d, odivs)) and f not in vcos:
                            if f not in vcos:
                                vcos.append(f)
                                config = {
                                    "n2": n,
                                    "r2": r,
                                    "vco": f,
                                    "required_output_divs": d,
                                }
                                configs.append(config)
                                if len(configs) >= find:
                                    return configs

        return configs



"""Session state management for Streamlit application."""

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

from streamlit.runtime import get_instance
from streamlit.runtime.legacy_caching.hashing import _CodeHasher
from streamlit.runtime.scriptrunner import get_script_run_ctx

if TYPE_CHECKING:
    from streamlit.runtime.state.session_state import SessionState as Session


class _SessionState:
    """Internal session state management class."""

    def __init__(
        self, session: "Session", hash_funcs: Optional[Dict[type, Callable[[Any], str]]]
    ) -> None:
        """Initialize SessionState instance."""
        self.__dict__["_state"] = {
            "data": {},
            "hash": None,
            "hasher": _CodeHasher(hash_funcs),
            "is_rerun": False,
            "session": session,
        }

    def __call__(self, **kwargs: Dict[str, Any]) -> None:  # type: ignore[misc]
        """Initialize state data once."""
        for item, value in kwargs.items():
            if item not in self._state["data"]:
                self._state["data"][item] = value

    def __getitem__(self, item: str) -> Any:  # noqa: ANN401
        """Return a saved state value, None if item is undefined."""
        return self._state["data"].get(item, None)

    def __getattr__(self, item: str) -> Any:  # noqa: ANN401
        """Return a saved state value, None if item is undefined."""
        return self._state["data"].get(item, None)

    def __setitem__(self, item: str, value: Any) -> None:  # noqa: ANN401
        """Set state value."""
        self._state["data"][item] = value

    def __setattr__(self, item: str, value: Any) -> None:  # noqa: ANN401
        """Set state value."""
        self._state["data"][item] = value

    def clear(self) -> None:
        """Clear session state and request a rerun."""
        self._state["data"].clear()
        self._state["session"].request_rerun()

    def sync(self) -> None:
        """Rerun the app with all state values up to date.

        This fixes rollbacks from the beginning.
        """
        # Ensure to rerun only once to avoid infinite loops
        # caused by a constantly changing state value at each run.
        #
        # Example: state.value += 1
        if self._state["is_rerun"]:
            self._state["is_rerun"] = False

        elif self._state["hash"] is not None:
            if self._state["hash"] != self._state["hasher"].to_bytes(
                self._state["data"], None
            ):
                self._state["is_rerun"] = True
                self._state["session"].request_rerun(None)

        self._state["hash"] = self._state["hasher"].to_bytes(self._state["data"], None)


def _get_session() -> Any:  # noqa: ANN401
    """Get the current Streamlit session."""
    runtime = get_instance()
    session_id = get_script_run_ctx().session_id
    session_info = runtime._session_mgr.get_session_info(session_id)

    if session_info is None:
        raise RuntimeError("Couldn't get your Streamlit Session object.")

    return session_info.session


def get_state(
    hash_funcs: Optional[Dict[type, Callable[[Any], str]]] = None,
) -> _SessionState:
    """Get or create session state.

    Args:
        hash_funcs: Optional hash functions for state comparison

    Returns:
        Session state object
    """
    session = _get_session()

    if not hasattr(session, "_custom_session_state"):
        session._custom_session_state = _SessionState(session, hash_funcs)

    return session._custom_session_state


# Only used for separating namespace
# Everything can be saved at state variable as well.
CONFIG_DEFAULTS: Dict[str, Any] = {"slider_value": 0}


def provide_state(
    hash_funcs: Optional[Dict[type, Callable[[Any], str]]] = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to provide state to a function.

    Args:
        hash_funcs: Optional hash functions for state comparison

    Returns:
        Decorator function
    """

    def inner(func: Callable[..., Any]) -> Callable[..., Any]:  # noqa: ANN401
        """Inner decorator function."""

        def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
            """Wrapper function that injects state."""
            state = get_state(hash_funcs=hash_funcs)
            if state.client_config is None:
                state.client_config = deepcopy(CONFIG_DEFAULTS)

            return_value = func(*args, state=state, **kwargs)  # noqa: B026
            state.sync()
            return return_value

        return wrapper

    return inner
